#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import neo4j
import pytest
from neo4j_graphrag.experimental.components.kg_writer import Neo4jWriter
from neo4j_graphrag.experimental.components.lexical_graph import LexicalGraphBuilder
from neo4j_graphrag.experimental.components.types import (
    LexicalGraphConfig,
    TextChunk,
    TextChunks,
)
from neo4j_graphrag.experimental.pipeline import Pipeline


@pytest.mark.asyncio
@pytest.mark.usefixtures("setup_neo4j_for_kg_construction")
async def test_lexical_graph_component_alone_default_config(
    driver: neo4j.Driver,
) -> None:
    pipe = Pipeline()
    pipe.add_component(LexicalGraphBuilder(), "lexical_graph")
    pipe.add_component(Neo4jWriter(driver), "writer")
    pipe.connect("lexical_graph", "writer", {"graph": "lexical_graph.graph"})

    result = await pipe.run(
        {
            "lexical_graph": {
                "text_chunks": TextChunks(chunks=[TextChunk(text="my text", index=0)])
            }
        }
    )
    assert result.result == {
        "writer": {
            "status": "SUCCESS",
            "metadata": {"node_count": 1, "relationship_count": 0},
        }
    }
    default_config = LexicalGraphConfig()
    created_chunks = driver.execute_query(
        f"MATCH (n:{default_config.chunk_node_label}) RETURN n"
    )
    assert len(created_chunks.records) == 1
    created_chunks = driver.execute_query(
        f"MATCH (n:{default_config.document_node_label}) RETURN n"
    )
    assert len(created_chunks.records) == 0
    created_rels = driver.execute_query("MATCH ()-[r]->() RETURN r")
    assert len(created_rels.records) == 0



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

import warnings
from typing import Any, Optional

from neo4j_graphrag.exceptions import (
    PromptMissingInputError,
    PromptMissingPlaceholderError,
)


class PromptTemplate:
    """This class is used to generate a parameterized prompt. It is defined
    from a string (the template) using the Python format syntax (parameters
    between curly braces `{}`) and a list of required inputs.
    Before sending the instructions to an LLM, call the `format` method that will
    replace parameters with the provided values. If any of the expected inputs is
    missing, a `PromptMissingInputError` is raised.
    """

    DEFAULT_SYSTEM_INSTRUCTIONS: str = ""
    DEFAULT_TEMPLATE: str = ""
    EXPECTED_INPUTS: list[str] = list()

    def __init__(
        self,
        template: Optional[str] = None,
        expected_inputs: Optional[list[str]] = None,
        system_instructions: Optional[str] = None,
    ) -> None:
        self.template = template or self.DEFAULT_TEMPLATE
        self.expected_inputs = expected_inputs or self.EXPECTED_INPUTS
        self.system_instructions = (
            system_instructions or self.DEFAULT_SYSTEM_INSTRUCTIONS
        )

        for e in self.expected_inputs:
            if f"{{{e}}}" not in self.template:
                raise PromptMissingPlaceholderError(
                    f"`template` is missing placeholder {e}"
                )

    def _format(self, **kwargs: Any) -> str:
        for e in self.EXPECTED_INPUTS:
            if e not in kwargs:
                raise PromptMissingInputError(f"Missing input '{e}'")
        return self.template.format(**kwargs)

    def format(self, *args: Any, **kwargs: Any) -> str:
        """This method is used to replace parameters with the provided values.
        Parameters must be provided:
        - as kwargs
        - as args if using the same order as in the expected inputs

        Example:

        .. code-block:: python

            prompt_template = PromptTemplate(
                template='''Explain the following concept to {target_audience}:
                Concept: {concept}
                Answer:
                ''',
                expected_inputs=['target_audience', 'concept']
            )
            prompt = prompt_template.format('12 yo children', concept='graph database')
            print(prompt)

            # Result:
            # '''Explain the following concept to 12 yo children:
            # Concept: graph database
            # Answer:
            # '''

        """
        data = kwargs
        data.update({k: v for k, v in zip(self.expected_inputs, args)})
        return self._format(**data)


class RagTemplate(PromptTemplate):
    DEFAULT_SYSTEM_INSTRUCTIONS = "Answer the user question using the provided context."
    DEFAULT_TEMPLATE = """Context:
{context}

Examples:
{examples}

Question:
{query_text}

Answer:
"""
    EXPECTED_INPUTS = ["context", "query_text", "examples"]

    def format(self, query_text: str, context: str, examples: str) -> str:
        return super().format(query_text=query_text, context=context, examples=examples)


class Text2CypherTemplate(PromptTemplate):
    DEFAULT_TEMPLATE = """
Task: Generate a Cypher statement for querying a Neo4j graph database from a user input.

Schema:
{schema}

Examples (optional):
{examples}

Input:
{query_text}

Do not use any properties or relationships not included in the schema.
Do not include triple backticks ``` or any additional text except the generated Cypher statement in your response.

Cypher query:
"""
    EXPECTED_INPUTS = ["query_text"]

    def format(
        self,
        schema: Optional[str] = None,
        examples: Optional[str] = None,
        query_text: str = "",
        query: Optional[str] = None,
        **kwargs: Any,
    ) -> str:
        if query is not None:
            if query_text:
                warnings.warn(
                    "Both 'query' and 'query_text' are provided, 'query_text' will be used.",
                    DeprecationWarning,
                    stacklevel=2,
                )
            elif isinstance(query, str):
                warnings.warn(
                    "'query' is deprecated and will be removed in a future version, please use 'query_text' instead.",
                    DeprecationWarning,
                    stacklevel=2,
                )
                query_text = query

        return super().format(
            query_text=query_text, schema=schema, examples=examples, **kwargs
        )


class ERExtractionTemplate(PromptTemplate):
    DEFAULT_TEMPLATE = """
You are a top-tier algorithm designed for extracting
information in structured formats to build a knowledge graph.

Extract the entities (nodes) and specify their type from the following text.
Also extract the relationships between these nodes.

Return result as JSON using the following format:
{{"nodes": [ {{"id": "0", "label": "Person", "properties": {{"name": "John"}} }}],
"relationships": [{{"type": "KNOWS", "start_node_id": "0", "end_node_id": "1", "properties": {{"since": "2024-08-01"}} }}] }}

Use only the following node and relationship types (if provided):
{schema}

Assign a unique ID (string) to each node, and reuse it to define relationships.
Do respect the source and target node types for relationship and
the relationship direction.

Make sure you adhere to the following rules to produce valid JSON objects:
- Do not return any additional information other than the JSON in it.
- Omit any backticks around the JSON - simply output the JSON on its own.
- The JSON object must not wrapped into a list - it is its own JSON object.
- Property names must be enclosed in double quotes

Examples:
{examples}

Input text:

{text}
"""
    EXPECTED_INPUTS = ["text"]

    def format(
        self,
        schema: dict[str, Any],
        examples: str,
        text: str = "",
    ) -> str:
        return super().format(text=text, schema=schema, examples=examples)


class SchemaExtractionTemplate(PromptTemplate):
    DEFAULT_TEMPLATE = """
You are a top-tier algorithm designed for extracting a labeled property graph schema in
structured formats.

Generate a generalized graph schema based on the input text. Identify key node types,
their relationship types, and property types.

IMPORTANT RULES:
1. Return only abstract schema information, not concrete instances.
2. Use singular PascalCase labels for node types (e.g., Person, Company, Product).
3. Use UPPER_SNAKE_CASE labels for relationship types (e.g., WORKS_FOR, MANAGES).
4. Include property definitions only when the type can be confidently inferred, otherwise omit them.
5. When defining patterns, ensure that every node label and relationship label mentioned exists in your lists of node types and relationship types.
6. Do not create node types that aren't clearly mentioned in the text.
7. Keep your schema minimal and focused on clearly identifiable patterns in the text.
8. UNIQUENESS CONSTRAINTS:
8.1 UNIQUENESS is optional; each node_type may or may not have exactly one uniqueness constraint.
8.2 Only use properties that seem to not have too many missing values in the sample.
8.3 Constraints reference node_types by label and specify which property is unique.
8.4 If a property appears in a uniqueness constraint it MUST also appear in the corresponding node_type as a property.
9. REQUIRED PROPERTIES:
9.1 Mark a property as "required": true if every instance of that node/relationship type MUST have this property (non-nullable).
9.2 Mark a property as "required": false if the property is optional and may be absent on some instances.
9.3 Properties that are identifiers, names, or essential characteristics are typically required.
9.4 Properties that are supplementary information (phone numbers, descriptions, metadata) are typically optional.
9.5 When uncertain, default to "required": false.
9.6 If a property has a UNIQUENESS constraint, it MUST be marked as "required": true.

Accepted property types are: BOOLEAN, DATE, DURATION, FLOAT, INTEGER, LIST,
LOCAL_DATETIME, LOCAL_TIME, POINT, STRING, ZONED_DATETIME, ZONED_TIME.

Return a valid JSON object that follows this precise structure:
{{
  "node_types": [
    {{
      "label": "Person",
      "properties": [
        {{
          "name": "name",
          "type": "STRING",
          "required": true
        }},
        {{
          "name": "email",
          "type": "STRING",
          "required": false
        }}
      ]
    }}
    ...
  ],
  "relationship_types": [
    {{
      "label": "WORKS_FOR"
    }}
    ...
  ],
  "patterns": [
    {{"source": "Person", "relationship": "WORKS_FOR", "target": "Company"}},
    ...
  ],
  "constraints": [
    {{
      "type": "UNIQUENESS",
      "node_type": "Person",
      "property_name": "name"
    }}
    ...
  ]
}}

Examples:
{examples}

Input text:
{text}
"""
    EXPECTED_INPUTS = ["text"]

    def format(
        self,
        text: str = "",
        examples: str = "",
    ) -> str:
        return super().format(text=text, examples=examples)



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from unittest.mock import Mock, patch

import pytest
from neo4j_graphrag.experimental.pipeline import (
    Component,
    Pipeline,
)
from neo4j_graphrag.experimental.pipeline.exceptions import (
    PipelineDefinitionError,
    PipelineMissingDependencyError,
    PipelineStatusUpdateError,
)
from neo4j_graphrag.experimental.pipeline.orchestrator import Orchestrator
from neo4j_graphrag.experimental.pipeline.types.orchestration import RunStatus

from tests.unit.experimental.pipeline.components import (
    ComponentNoParam,
    ComponentPassThrough,
)


def test_orchestrator_get_input_config_for_task_pipeline_not_validated() -> None:
    pipe = Pipeline()
    pipe.add_component(ComponentPassThrough(), "a")
    pipe.add_component(ComponentPassThrough(), "b")
    orchestrator = Orchestrator(pipe)
    with pytest.raises(PipelineDefinitionError) as exc:
        orchestrator.get_input_config_for_task(pipe.get_node_by_name("a"))
    assert "You must validate the pipeline input config first" in str(exc.value)


@pytest.mark.asyncio
async def test_orchestrator_get_component_inputs_from_user_only() -> None:
    """Components take all their inputs from user input."""
    pipe = Pipeline()
    pipe.add_component(ComponentPassThrough(), "a")
    pipe.add_component(ComponentPassThrough(), "b")
    orchestrator = Orchestrator(pipe)
    input_data = {
        "a": {"value": "user input for component a"},
        "b": {"value": "user input for component b"},
    }
    data = await orchestrator.get_component_inputs("a", {}, input_data)
    assert data == {"value": "user input for component a"}
    data = await orchestrator.get_component_inputs("b", {}, input_data)
    assert data == {"value": "user input for component b"}


@patch(
    "neo4j_graphrag.experimental.pipeline.pipeline.Orchestrator.get_results_for_component"
)
@pytest.mark.asyncio
async def test_orchestrator_get_component_inputs_from_parent_specific(
    mock_result: Mock,
) -> None:
    """Propagate one specific output field from parent to a child component."""
    pipe = Pipeline()
    pipe.add_component(ComponentPassThrough(), "a")
    pipe.add_component(ComponentPassThrough(), "b")
    pipe.connect("a", "b", input_config={"value": "a.result"})

    # component "a" already run and results stored:
    mock_result.return_value = {"result": "output from component a"}

    orchestrator = Orchestrator(pipe)
    data = await orchestrator.get_component_inputs(
        "b", {"value": {"component": "a", "param": "result"}}, {}
    )
    assert data == {"value": "output from component a"}


@patch(
    "neo4j_graphrag.experimental.pipeline.pipeline.Orchestrator.get_results_for_component"
)
@pytest.mark.asyncio
async def test_orchestrator_get_component_inputs_from_parent_all(
    mock_result: Mock,
) -> None:
    """Use the component name to get the full output
    (without extracting a specific field).
    """
    pipe = Pipeline()
    pipe.add_component(ComponentNoParam(), "a")
    pipe.add_component(ComponentPassThrough(), "b")
    pipe.connect("a", "b", input_config={"value": "a"})

    # component "a" already run and results stored:
    mock_result.return_value = {"result": "output from component a"}

    orchestrator = Orchestrator(pipe)
    data = await orchestrator.get_component_inputs(
        "b", {"value": {"component": "a"}}, {}
    )
    assert data == {"value": {"result": "output from component a"}}


@patch(
    "neo4j_graphrag.experimental.pipeline.pipeline.Orchestrator.get_results_for_component"
)
@pytest.mark.asyncio
async def test_orchestrator_get_component_inputs_from_parent_and_input(
    mock_result: Mock,
) -> None:
    """Some parameters from user input, some other parameter from previous component."""
    pipe = Pipeline()
    pipe.add_component(ComponentNoParam(), "a")
    pipe.add_component(ComponentPassThrough(), "b")
    pipe.connect("a", "b", input_config={"value": "a"})

    # component "a" already run and results stored:
    mock_result.return_value = {"result": "output from component a"}

    orchestrator = Orchestrator(pipe)
    data = await orchestrator.get_component_inputs(
        "b",
        {"value": {"component": "a"}},
        {"b": {"other_value": "user input for component b 'other_value' param"}},
    )
    assert data == {
        "value": {"result": "output from component a"},
        "other_value": "user input for component b 'other_value' param",
    }


@patch(
    "neo4j_graphrag.experimental.pipeline.pipeline.Orchestrator.get_results_for_component"
)
@pytest.mark.asyncio
async def test_orchestrator_get_component_inputs_ignore_user_input_if_input_def_provided(
    mock_result: Mock,
) -> None:
    """If a parameter is defined both in the user input and in an input definition
    (ie propagated from a previous component), the user input is ignored and a
    warning is raised.
    """
    pipe = Pipeline()
    pipe.add_component(ComponentNoParam(), "a")
    pipe.add_component(ComponentPassThrough(), "b")
    pipe.connect("a", "b", input_config={"value": "a"})

    # component "a" already run and results stored:
    mock_result.return_value = {"result": "output from component a"}

    orchestrator = Orchestrator(pipe)
    with pytest.warns(Warning) as w:
        data = await orchestrator.get_component_inputs(
            "b",
            {"value": {"component": "a"}},
            {"b": {"value": "user input for component a"}},
        )
        assert data == {"value": {"result": "output from component a"}}
        assert (
            w[0].message.args[0]  # type: ignore[union-attr]
            == "In component 'b', parameter 'value' from user input will be ignored and replaced by 'a.value'"
        )


@pytest.mark.asyncio
@patch(
    "neo4j_graphrag.experimental.pipeline.pipeline.Orchestrator.get_status_for_component"
)
@pytest.mark.parametrize(
    "old_status, new_status, result",
    [
        # Normal path: from UNKNOWN to RUNNING to DONE
        (RunStatus.UNKNOWN, RunStatus.RUNNING, "ok"),
        (RunStatus.RUNNING, RunStatus.DONE, "ok"),
        # Error: status is already set to this value
        (RunStatus.RUNNING, RunStatus.RUNNING, "Status is already RunStatus.RUNNING"),
        (RunStatus.DONE, RunStatus.DONE, "Status is already RunStatus.DONE"),
        # Error: can't go back in time
        (
            RunStatus.DONE,
            RunStatus.RUNNING,
            "Can't go from RunStatus.DONE to RunStatus.RUNNING",
        ),
        (
            RunStatus.RUNNING,
            RunStatus.UNKNOWN,
            "Can't go from RunStatus.RUNNING to RunStatus.UNKNOWN",
        ),
        (
            RunStatus.DONE,
            RunStatus.UNKNOWN,
            "Can't go from RunStatus.DONE to RunStatus.UNKNOWN",
        ),
    ],
)
async def test_orchestrator_set_component_status(
    mock_status: Mock,
    old_status: RunStatus,
    new_status: RunStatus,
    result: str,
) -> None:
    pipe = Pipeline()
    orchestrator = Orchestrator(pipeline=pipe)
    mock_status.side_effect = [
        old_status,
    ]
    if result == "ok":
        await orchestrator.set_task_status("task_name", new_status)
    else:
        with pytest.raises(PipelineStatusUpdateError) as exc:
            await orchestrator.set_task_status("task_name", new_status)
        assert result in str(exc)


@pytest.fixture(scope="function")
def pipeline_branch() -> Pipeline:
    pipe = Pipeline()
    pipe.add_component(Component(), "a")  # type: ignore[abstract,unused-ignore]
    pipe.add_component(Component(), "b")  # type: ignore[abstract,unused-ignore]
    pipe.add_component(Component(), "c")  # type: ignore[abstract,unused-ignore]
    pipe.connect("a", "b")
    pipe.connect("a", "c")
    return pipe


@pytest.fixture(scope="function")
def pipeline_aggregation() -> Pipeline:
    pipe = Pipeline()
    pipe.add_component(Component(), "a")  # type: ignore[abstract,unused-ignore]
    pipe.add_component(Component(), "b")  # type: ignore[abstract,unused-ignore]
    pipe.add_component(Component(), "c")  # type: ignore[abstract,unused-ignore]
    pipe.connect("a", "c")
    pipe.connect("b", "c")
    return pipe


@pytest.mark.asyncio
@patch(
    "neo4j_graphrag.experimental.pipeline.pipeline.Orchestrator.get_status_for_component"
)
async def test_orchestrator_check_dependency_complete(
    mock_status: Mock, pipeline_branch: Pipeline
) -> None:
    """a -> b, c"""
    orchestrator = Orchestrator(pipeline=pipeline_branch)
    node_a = pipeline_branch.get_node_by_name("a")
    await orchestrator.check_dependencies_complete(node_a)
    node_b = pipeline_branch.get_node_by_name("b")
    # dependency is DONE:
    mock_status.side_effect = [RunStatus.DONE]
    await orchestrator.check_dependencies_complete(node_b)
    # dependency is not DONE:
    mock_status.side_effect = [RunStatus.RUNNING]
    with pytest.raises(PipelineMissingDependencyError):
        await orchestrator.check_dependencies_complete(node_b)


@pytest.mark.asyncio
@patch(
    "neo4j_graphrag.experimental.pipeline.pipeline.Orchestrator.get_status_for_component"
)
@patch(
    "neo4j_graphrag.experimental.pipeline.pipeline.Orchestrator.check_dependencies_complete",
)
async def test_orchestrator_next_task_branch_no_missing_dependencies(
    mock_dep: Mock, mock_status: Mock, pipeline_branch: Pipeline
) -> None:
    """a -> b, c"""
    orchestrator = Orchestrator(pipeline=pipeline_branch)
    node_a = pipeline_branch.get_node_by_name("a")
    mock_status.side_effect = [
        # next "b"
        RunStatus.UNKNOWN,
        # next "c"
        RunStatus.UNKNOWN,
    ]
    mock_dep.side_effect = [
        None,  # "b" has no missing dependencies
        None,  # "c" has no missing dependencies
    ]
    next_tasks = [n async for n in orchestrator.next(node_a)]
    next_task_names = [n.name for n in next_tasks]
    assert next_task_names == ["b", "c"]


@pytest.mark.asyncio
@patch(
    "neo4j_graphrag.experimental.pipeline.pipeline.Orchestrator.get_status_for_component"
)
@patch(
    "neo4j_graphrag.experimental.pipeline.pipeline.Orchestrator.check_dependencies_complete",
)
async def test_orchestrator_next_task_branch_missing_dependencies(
    mock_dep: Mock, mock_status: Mock, pipeline_branch: Pipeline
) -> None:
    """a -> b, c"""
    orchestrator = Orchestrator(pipeline=pipeline_branch)
    node_a = pipeline_branch.get_node_by_name("a")
    mock_status.side_effect = [
        # next "b"
        RunStatus.UNKNOWN,
        # next "c"
        RunStatus.UNKNOWN,
    ]
    mock_dep.side_effect = [
        PipelineMissingDependencyError,  # "b" has missing dependencies
        None,  # "c" has no missing dependencies
    ]
    next_tasks = [n async for n in orchestrator.next(node_a)]
    next_task_names = [n.name for n in next_tasks]
    assert next_task_names == ["c"]


@pytest.mark.asyncio
@patch(
    "neo4j_graphrag.experimental.pipeline.pipeline.Orchestrator.get_status_for_component"
)
@patch(
    "neo4j_graphrag.experimental.pipeline.pipeline.Orchestrator.check_dependencies_complete",
)
async def test_orchestrator_next_task_aggregation_no_missing_dependencies(
    mock_dep: Mock, mock_status: Mock, pipeline_aggregation: Pipeline
) -> None:
    """a, b -> c"""
    orchestrator = Orchestrator(pipeline=pipeline_aggregation)
    node_a = pipeline_aggregation.get_node_by_name("a")
    mock_status.side_effect = [
        RunStatus.UNKNOWN,  # status for "c", not started
    ]
    mock_dep.side_effect = [
        None,  # no missing deps
    ]
    # then "c" can start
    next_tasks = [n async for n in orchestrator.next(node_a)]
    next_task_names = [n.name for n in next_tasks]
    assert next_task_names == ["c"]


@pytest.mark.asyncio
@patch(
    "neo4j_graphrag.experimental.pipeline.pipeline.Orchestrator.get_status_for_component"
)
@patch(
    "neo4j_graphrag.experimental.pipeline.pipeline.Orchestrator.check_dependencies_complete",
)
async def test_orchestrator_next_task_aggregation_missing_dependency(
    mock_dep: Mock, mock_status: Mock, pipeline_aggregation: Pipeline
) -> None:
    """a, b -> c"""
    orchestrator = Orchestrator(pipeline=pipeline_aggregation)
    node_a = pipeline_aggregation.get_node_by_name("a")
    mock_status.side_effect = [
        RunStatus.UNKNOWN,  # status for "c" is unknown, it's a possible next
    ]
    mock_dep.side_effect = [
        PipelineMissingDependencyError,  # some dependencies are not done yet
    ]
    next_task_names = [n.name async for n in orchestrator.next(node_a)]
    # "c" dependencies not ready yet
    assert next_task_names == []


@pytest.mark.asyncio
@patch(
    "neo4j_graphrag.experimental.pipeline.pipeline.Orchestrator.get_status_for_component"
)
async def test_orchestrator_next_task_aggregation_next_already_started(
    mock_status: Mock, pipeline_aggregation: Pipeline
) -> None:
    """a, b -> c"""
    orchestrator = Orchestrator(pipeline=pipeline_aggregation)
    node_a = pipeline_aggregation.get_node_by_name("a")
    mock_status.side_effect = [
        RunStatus.RUNNING,  # status for "c" is already running, do not start it again
    ]
    next_task_names = [n.name async for n in orchestrator.next(node_a)]
    assert next_task_names == []



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging

import neo4j
import pytest
from neo4j_graphrag.experimental.components.kg_writer import Neo4jWriter
from neo4j_graphrag.experimental.components.types import (
    Neo4jGraph,
    Neo4jNode,
    Neo4jRelationship,
)


@pytest.mark.asyncio
@pytest.mark.usefixtures("setup_neo4j_for_kg_construction")
async def test_kg_writer(driver: neo4j.Driver) -> None:
    start_node = Neo4jNode(
        id="1",
        label="MyLabel",
        properties={"id": "abc"},
        embedding_properties={"vectorProperty": [1.0, 2.0, 3.0]},
    )
    end_node = Neo4jNode(
        id="2",
        label="MyLabel",
        properties={"id": "def"},
    )
    relationship = Neo4jRelationship(
        start_node_id="1", end_node_id="2", type="MY_RELATIONSHIP"
    )
    node_with_two_embeddings = Neo4jNode(
        id="3",
        label="MyLabel",
        properties={"id": "ghi"},
        embedding_properties={
            "vectorProperty": [1.0, 2.0, 3.0],
            "otherVectorProperty": [10.0, 20.0, 30.0],
        },
    )
    graph = Neo4jGraph(
        nodes=[start_node, end_node, node_with_two_embeddings],
        relationships=[relationship],
    )

    neo4j_writer = Neo4jWriter(driver=driver)
    res = await neo4j_writer.run(graph=graph)
    assert res.status == "SUCCESS"

    query = """
    MATCH (a:MyLabel {id: 'abc'})-[r:MY_RELATIONSHIP]->(b:MyLabel {id: 'def'})
    RETURN a, r, b
    """
    record = driver.execute_query(query).records[0]
    assert "a" and "b" and "r" in record.keys()

    node_a = record["a"]
    assert start_node.label in list(node_a.labels)
    assert start_node.properties.get("id") == str(node_a.get("id"))
    for key, val in start_node.properties.items():
        assert key in node_a.keys()
        assert val == node_a.get(key)
    if start_node.embedding_properties:  # for mypy
        for emb_key, emb_val in start_node.embedding_properties.items():
            assert emb_key in node_a.keys()
            assert emb_val == node_a.get(emb_key)

    node_b = record["b"]
    assert end_node.label in list(node_b.labels)
    assert end_node.properties.get("id") == str(node_b.get("id"))
    for key, val in end_node.properties.items():
        assert key in node_b.keys()
        assert val == node_b.get(key)

    rel = record["r"]
    assert rel.type == relationship.type
    assert rel.start_node.get("id") == start_node.properties.get("id")
    assert rel.end_node.get("id") == end_node.properties.get("id")

    query = """
    MATCH (c:MyLabel {id: 'ghi'})
    RETURN c
    """
    records = driver.execute_query(query).records
    assert len(records) == 1
    node_c = records[0]["c"]
    if node_with_two_embeddings.embedding_properties:  # for mypy
        for emb_key, emb_val in node_with_two_embeddings.embedding_properties.items():
            assert emb_key in node_c.keys()
            assert emb_val == node_c.get(emb_key)


@pytest.mark.asyncio
@pytest.mark.usefixtures("setup_neo4j_for_kg_construction")
async def test_kg_writer_no_neo4j_deprecation_warning(
    driver: neo4j.Driver, caplog: pytest.LogCaptureFixture
) -> None:
    start_node = Neo4jNode(
        id="1",
        label="MyLabel",
        properties={"chunk": 1},
        embedding_properties={"vectorProperty": [1.0, 2.0, 3.0]},
    )
    end_node = Neo4jNode(
        id="2",
        label="MyLabel",
        properties={},
    )
    relationship = Neo4jRelationship(
        start_node_id="1", end_node_id="2", type="MY_RELATIONSHIP"
    )
    graph = Neo4jGraph(
        nodes=[start_node, end_node],
        relationships=[relationship],
    )

    neo4j_writer = Neo4jWriter(driver=driver)
    with caplog.at_level(logging.WARNING):
        res = await neo4j_writer.run(graph=graph)

    for record in caplog.records:
        if (
            "Neo.ClientNotification.Statement.FeatureDeprecationWarning"
            in record.message
        ):
            assert False, f"Deprecation warning found in logs: {record.message}"

    assert res.status == "SUCCESS"



"""End to end example of building a RAG pipeline backed by a Neo4j database.
Requires OPENAI_API_KEY to be in the env var.

This example illustrates:
- VectorCypherRetriever with a custom formatter function to extract relevant
    context from neo4j result
- Use of a custom prompt for RAG
- Logging configuration
"""

import neo4j
from neo4j_graphrag.embeddings.openai import OpenAIEmbeddings
from neo4j_graphrag.generation import GraphRAG, RagTemplate
from neo4j_graphrag.llm import OpenAILLM
from neo4j_graphrag.retrievers import VectorCypherRetriever

URI = "neo4j://localhost:7687"
AUTH = ("neo4j", "password")
DATABASE = "neo4j"
INDEX = "moviePlotsEmbedding"


driver = neo4j.GraphDatabase.driver(
    URI,
    auth=AUTH,
)

embedder = OpenAIEmbeddings()

retriever = VectorCypherRetriever(
    driver,
    index_name=INDEX,
    retrieval_query="WITH node, score RETURN node.title as title, node.plot as plot",
    embedder=embedder,
    neo4j_database=DATABASE,
)

llm = OpenAILLM(model_name="gpt-4o", model_params={"temperature": 0})

template = RagTemplate(
    template="""You are an expert at movies and actors. Your task is to
    answer the user's question based on the provided context. Use only the
    information within that context.

    Context:
    {context}

    Question:
    {query_text}

    Answer:
    """
)

rag = GraphRAG(retriever=retriever, llm=llm, prompt_template=template)

result = rag.search(
    "Tell me more about Avatar movies",
    return_context=True,
)
print(result.answer)

driver.close()



#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# built-in dependencies
from __future__ import annotations

import inspect
import logging
from typing import Any, List, Optional, Sequence, Type, Union, cast, overload

# 3rd party dependencies
from pydantic import BaseModel, ValidationError

# project dependencies
from neo4j_graphrag.exceptions import LLMGenerationError
from neo4j_graphrag.llm.base import LLMInterface, LLMInterfaceV2
from neo4j_graphrag.llm.types import (
    BaseMessage,
    LLMResponse,
    MessageList,
    ToolCall,
    ToolCallResponse,
)
from neo4j_graphrag.message_history import MessageHistory
from neo4j_graphrag.tool import Tool
from neo4j_graphrag.types import LLMMessage
from neo4j_graphrag.utils.rate_limit import (
    RateLimitHandler,
)
from neo4j_graphrag.utils.rate_limit import (
    async_rate_limit_handler as async_rate_limit_handler_decorator,
)
from neo4j_graphrag.utils.rate_limit import (
    rate_limit_handler as rate_limit_handler_decorator,
)

try:
    from vertexai.generative_models import (
        Content,
        FunctionCall,
        FunctionDeclaration,
        GenerationResponse,
        GenerativeModel,
        Part,
        ResponseValidationError,
        ToolConfig,
    )
    from vertexai.generative_models import (
        Tool as VertexAITool,
    )
except ImportError:
    GenerativeModel = None  # type: ignore[misc, assignment]
    ResponseValidationError = None  # type: ignore[misc, assignment]

logger = logging.getLogger(__name__)

# Params to exclude when extracting from GenerationConfig for structured output
_GENERATION_CONFIG_SCHEMA_PARAMS = {"response_schema", "response_mime_type"}


def _extract_generation_config_params(
    config: Any, exclude_schema: bool = True
) -> dict[str, Any]:
    """Extract valid parameters from a GenerationConfig object.

    This function extracts parameters from the internal _raw_generation_config
    protobuf and returns them as a dict that can be passed to GenerationConfig().

    Args:
        config: A GenerationConfig object
        exclude_schema: If True, excludes response_schema and response_mime_type

    Returns:
        Dict of parameter name to value for non-empty params
    """
    from vertexai.generative_models import GenerationConfig

    if not hasattr(config, "_raw_generation_config"):
        return {}

    raw = config._raw_generation_config

    # Get valid params from GenerationConfig signature
    sig = inspect.signature(GenerationConfig.__init__)
    valid_params = {
        name
        for name, _ in sig.parameters.items()
        if name != "self"
        and (not exclude_schema or name not in _GENERATION_CONFIG_SCHEMA_PARAMS)
    }

    preserved = {}
    for param in valid_params:
        val = getattr(raw, param, None)
        if val:  # Only include non-empty values
            # Convert repeated fields (like stop_sequences) to lists
            if hasattr(val, "__iter__") and not isinstance(val, (str, bytes, dict)):
                val = list(val)
            preserved[param] = val

    return preserved


# pylint: disable=arguments-differ, redefined-builtin, no-else-return
class VertexAILLM(LLMInterface, LLMInterfaceV2):
    """Interface for large language models on Vertex AI

    Args:
        model_name (str, optional): Name of the LLM to use. Defaults to "gemini-1.5-flash-001".
        model_params (Optional[dict], optional): Additional parameters for LLMInterface(V1) passed to the model when text is sent to it. Defaults to None.
        system_instruction: Optional[str], optional): Additional instructions for setting the behavior and context for the model in a conversation. Defaults to None.
        rate_limit_handler (Optional[RateLimitHandler], optional): Rate limit handler for LLMInterface(V1). Defaults to None.
        **kwargs (Any): Arguments passed to the model when for the class is initialised. Defaults to None.

    Raises:
        LLMGenerationError: If there's an error generating the response from the model.

    Example:

    .. code-block:: python

        from neo4j_graphrag.llm import VertexAILLM
        from vertexai.generative_models import GenerationConfig

        generation_config = GenerationConfig(temperature=0.0)
        llm = VertexAILLM(
            model_name="gemini-1.5-flash-001", generation_config=generation_config
        )
        llm.invoke("Who is the mother of Paul Atreides?")
    """

    supports_structured_output: bool = True

    def __init__(
        self,
        model_name: str = "gemini-1.5-flash-001",
        model_params: Optional[dict[str, Any]] = None,
        system_instruction: Optional[str] = None,
        rate_limit_handler: Optional[RateLimitHandler] = None,
        **kwargs: Any,
    ):
        if GenerativeModel is None or ResponseValidationError is None:
            raise ImportError(
                """Could not import Vertex AI Python client.
                Please install it with `pip install "neo4j-graphrag[google]"`."""
            )
            LLMInterfaceV2.__init__(
                self,
                model_name=model_name,
                model_params=model_params or {},
                rate_limit_handler=rate_limit_handler,
                **kwargs,
            )
        self.model_name = model_name
        self.system_instruction = system_instruction
        self.options = kwargs

    # overloads for LLMInterface and LLMInterfaceV2 methods
    @overload  # type: ignore[no-overload-impl]
    def invoke(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> LLMResponse: ...

    @overload
    def invoke(
        self,
        input: List[LLMMessage],
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse: ...

    @overload  # type: ignore[no-overload-impl]
    async def ainvoke(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> LLMResponse: ...

    @overload
    async def ainvoke(
        self,
        input: List[LLMMessage],
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse: ...

    # switching logics to LLMInterface or LLMInterfaceV2

    def invoke(  # type: ignore[no-redef]
        self,
        input: Union[str, List[LLMMessage]],
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse:
        if isinstance(input, str):
            return self.__invoke_v1(input, message_history, system_instruction)
        elif isinstance(input, list):
            return self.__invoke_v2(input, response_format=response_format, **kwargs)
        else:
            raise ValueError(f"Invalid input type for invoke method - {type(input)}")

    async def ainvoke(  # type: ignore[no-redef]
        self,
        input: Union[str, List[LLMMessage]],
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse:
        if isinstance(input, str):
            return await self.__ainvoke_v1(input, message_history, system_instruction)
        elif isinstance(input, list):
            return await self.__ainvoke_v2(
                input, response_format=response_format, **kwargs
            )
        else:
            raise ValueError(f"Invalid input type for ainvoke method - {type(input)}")

    def invoke_with_tools(
        self,
        input: str,
        tools: Sequence[Tool],  # Tools definition as a sequence of Tool objects
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> ToolCallResponse:
        return self.__invoke_v1_with_tools(
            input, tools, message_history, system_instruction
        )

    async def ainvoke_with_tools(
        self,
        input: str,
        tools: Sequence[Tool],
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> ToolCallResponse:
        return await self.__ainvoke_v1_with_tools(
            input, tools, message_history, system_instruction
        )

    # legacy and brand new implementations

    @rate_limit_handler_decorator
    def __invoke_v1(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> LLMResponse:
        """Sends text to the LLM and returns a response.

        Args:
            input (str): The text to send to the LLM.
            message_history (Optional[Union[List[LLMMessage], MessageHistory]]): A collection previous messages,
                with each message having a specific role assigned.
            system_instruction (Optional[str]): An option to override the llm system message for this invocation.

        Returns:
            LLMResponse: The response from the LLM.
        """
        model = self._get_model(
            system_instruction=system_instruction,
        )
        try:
            if isinstance(message_history, MessageHistory):
                message_history = message_history.messages
            options = self._get_call_params(input, message_history, tools=None)
            response = model.generate_content(**options)
            return self._parse_content_response(response)
        except ResponseValidationError as e:
            raise LLMGenerationError("Error calling VertexAILLM") from e

    def __invoke_v2(
        self,
        input: List[LLMMessage],
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse:
        """New invoke method for LLMInterfaceV2.

        Args:
            input (List[LLMMessage]): Input to the LLM.
            response_format (Optional[Union[Type[BaseModel], dict[str, Any]]]): Optional
                response format. Can be a Pydantic model class for structured output
                or a JSON schema dict.
            **kwargs: Additional parameters to pass to GenerationConfig (e.g., temperature,
                max_output_tokens, top_p, top_k). These override constructor values.

        Returns:
            LLMResponse: The response from the LLM.
        """
        system_instruction, messages = self.get_messages_v2(input)
        model = self._get_model(
            system_instruction=system_instruction,
        )
        try:
            options = self._get_call_params_v2(
                messages, tools=None, response_format=response_format, **kwargs
            )
            response = model.generate_content(**options)
            return self._parse_content_response(response)
        except ResponseValidationError as e:
            raise LLMGenerationError("Error calling VertexAILLM") from e

    @async_rate_limit_handler_decorator
    async def __ainvoke_v1(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> LLMResponse:
        """Asynchronously sends text to the LLM and returns a response.

        Args:
            input (str): The text to send to the LLM.
            message_history (Optional[Union[List[LLMMessage], MessageHistory]]): A collection previous messages,
                with each message having a specific role assigned.
            system_instruction (Optional[str]): An option to override the llm system message for this invocation.

        Returns:
            LLMResponse: The response from the LLM.
        """
        try:
            if isinstance(message_history, MessageHistory):
                message_history = message_history.messages
            model = self._get_model(
                system_instruction=system_instruction,
            )
            options = self._get_call_params(input, message_history, tools=None)
            response = await model.generate_content_async(**options)
            return self._parse_content_response(response)
        except ResponseValidationError as e:
            raise LLMGenerationError("Error calling VertexAILLM") from e

    async def __ainvoke_v2(
        self,
        input: list[LLMMessage],
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse:
        """Asynchronously sends text to the LLM and returns a response.

        Args:
            input (List[LLMMessage]): Input to the LLM.
            response_format (Optional[Union[Type[BaseModel], dict[str, Any]]]): Optional
                response format. Can be a Pydantic model class for structured output
                or a JSON schema dict.
            **kwargs: Additional parameters to pass to GenerationConfig (e.g., temperature,
                max_output_tokens, top_p, top_k). These override constructor values.

        Returns:
            LLMResponse: The response from the LLM.
        """
        try:
            system_instruction, messages = self.get_messages_v2(input)
            model = self._get_model(
                system_instruction=system_instruction,
            )
            options = self._get_call_params_v2(
                messages, tools=None, response_format=response_format, **kwargs
            )
            response = await model.generate_content_async(**options)
            return self._parse_content_response(response)
        except ResponseValidationError as e:
            raise LLMGenerationError("Error calling VertexAILLM") from e

    def __invoke_v1_with_tools(
        self,
        input: str,
        tools: Sequence[Tool],
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> ToolCallResponse:
        response = self._call_llm(
            input,
            message_history=message_history,
            system_instruction=system_instruction,
            tools=tools,
        )
        return self._parse_tool_response(response)

    async def __ainvoke_v1_with_tools(
        self,
        input: str,
        tools: Sequence[Tool],
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> ToolCallResponse:
        response = await self._acall_llm(
            input,
            message_history=message_history,
            system_instruction=system_instruction,
            tools=tools,
        )
        return self._parse_tool_response(response)

    # subsdiary methods

    def _to_vertexai_function_declaration(self, tool: Tool) -> FunctionDeclaration:
        return FunctionDeclaration(
            name=tool.get_name(),
            description=tool.get_description(),
            parameters=tool.get_parameters(exclude=["additional_properties"]),
        )

    def _get_llm_tools(
        self, tools: Optional[Sequence[Tool]]
    ) -> Optional[list[VertexAITool]]:
        if not tools:
            return None
        return [
            VertexAITool(
                function_declarations=[
                    self._to_vertexai_function_declaration(tool) for tool in tools
                ]
            )
        ]

    def _get_model(
        self,
        system_instruction: Optional[str] = None,
    ) -> GenerativeModel:
        # system_message = [system_instruction] if system_instruction is not None else []
        model = GenerativeModel(
            model_name=self.model_name,
            system_instruction=system_instruction,
        )
        return model

    def get_messages(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
    ) -> list[Content]:
        """Constructs messages for the Vertex AI model from input and message history."""
        messages = []
        if message_history:
            if isinstance(message_history, MessageHistory):
                message_history = message_history.messages
            try:
                MessageList(messages=cast(list[BaseMessage], message_history))
            except ValidationError as e:
                raise LLMGenerationError(e.errors()) from e

            for message in message_history:
                if message.get("role") == "user":
                    messages.append(
                        Content(
                            role="user",
                            parts=[Part.from_text(message.get("content", ""))],
                        )
                    )
                elif message.get("role") == "assistant":
                    messages.append(
                        Content(
                            role="model",
                            parts=[Part.from_text(message.get("content", ""))],
                        )
                    )

        messages.append(Content(role="user", parts=[Part.from_text(input)]))
        return messages

    def get_messages_v2(
        self,
        input: list[LLMMessage],
    ) -> tuple[str | None, list[Content]]:
        """Constructs messages for the Vertex AI model from input only."""
        messages = []
        system_instruction = self.system_instruction
        for message in input:
            role = message.get("role")
            if role == "system":
                system_instruction = message.get("content")
                continue
            if role == "user":
                messages.append(
                    Content(
                        role="user",
                        parts=[Part.from_text(message.get("content", ""))],
                    )
                )
                continue
            if role == "assistant":
                messages.append(
                    Content(
                        role="model",
                        parts=[Part.from_text(message.get("content", ""))],
                    )
                )
                continue
            raise ValueError(f"Unknown role: {role}")
        return system_instruction, messages

    def _get_call_params(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]],
        tools: Optional[Sequence[Tool]],
    ) -> dict[str, Any]:
        options = dict(self.options)
        if tools:
            # we want a tool back, remove generation_config if defined
            options.pop("generation_config", None)
            options["tools"] = self._get_llm_tools(tools)
            if "tool_config" not in options:
                options["tool_config"] = ToolConfig(
                    function_calling_config=ToolConfig.FunctionCallingConfig(
                        mode=ToolConfig.FunctionCallingConfig.Mode.ANY,
                    )
                )
        else:
            # no tools, remove tool_config if defined
            options.pop("tool_config", None)

        messages = self.get_messages(input, message_history)
        options["contents"] = messages
        return options

    def _get_call_params_v2(
        self,
        contents: list[Content],
        tools: Optional[Sequence[Tool]],
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        from vertexai.generative_models import GenerationConfig

        options = dict(self.options)
        if tools:
            # we want a tool back, remove generation_config if defined
            options.pop("generation_config", None)
            options["tools"] = self._get_llm_tools(tools)
            if "tool_config" not in options:
                options["tool_config"] = ToolConfig(
                    function_calling_config=ToolConfig.FunctionCallingConfig(
                        mode=ToolConfig.FunctionCallingConfig.Mode.ANY,
                    )
                )
        else:
            # no tools, remove tool_config if defined
            options.pop("tool_config", None)

            # Apply response_format and/or kwargs if provided
            if response_format is not None or kwargs:
                # Start with ALL existing params from constructor (including schema)
                existing_config = options.get("generation_config")
                params = _extract_generation_config_params(
                    existing_config, exclude_schema=False
                )

                # If response_format provided, override schema (prioritize it)
                if response_format is not None:
                    # Convert to JSON schema
                    if isinstance(response_format, type) and issubclass(
                        response_format, BaseModel
                    ):
                        # if we migrate to new google-genai-sdk, Pydantic models can be passed directly
                        schema = response_format.model_json_schema()
                    else:
                        schema = response_format
                    params["response_mime_type"] = "application/json"
                    params["response_schema"] = schema

                # Apply kwargs (they override constructor values but preserve schema)
                params.update(kwargs)

                options["generation_config"] = GenerationConfig(**params)
        options["contents"] = contents
        return options

    async def _acall_llm(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
        tools: Optional[Sequence[Tool]] = None,
    ) -> GenerationResponse:
        model = self._get_model(system_instruction=system_instruction)
        options = self._get_call_params(input, message_history, tools)
        response = await model.generate_content_async(**options)
        return response  # type: ignore[no-any-return]

    def _call_llm(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
        tools: Optional[Sequence[Tool]] = None,
    ) -> GenerationResponse:
        model = self._get_model(system_instruction=system_instruction)
        options = self._get_call_params(input, message_history, tools)
        response = model.generate_content(**options)
        return response  # type: ignore[no-any-return]

    def _to_tool_call(self, function_call: FunctionCall) -> ToolCall:
        return ToolCall(
            name=function_call.name,
            arguments=function_call.args,
        )

    def _parse_tool_response(self, response: GenerationResponse) -> ToolCallResponse:
        function_calls = response.candidates[0].function_calls
        return ToolCallResponse(
            tool_calls=[self._to_tool_call(f) for f in function_calls],
            content=None,
        )

    def _parse_content_response(self, response: GenerationResponse) -> LLMResponse:
        return LLMResponse(
            content=response.text,
        )



"""This example demonstrates how to use event callback to receive notifications
about the pipeline progress.
"""

from __future__ import annotations

import asyncio
import logging

import neo4j
from neo4j_graphrag.experimental.components.kg_writer import Neo4jWriter
from neo4j_graphrag.experimental.components.lexical_graph import LexicalGraphBuilder
from neo4j_graphrag.experimental.components.text_splitters.fixed_size_splitter import (
    FixedSizeSplitter,
)
from neo4j_graphrag.experimental.pipeline import Pipeline
from neo4j_graphrag.experimental.pipeline.pipeline import PipelineResult
from neo4j_graphrag.experimental.pipeline.notification import Event

logger = logging.getLogger(__name__)
logging.basicConfig()
logger.setLevel(logging.WARNING)


async def event_handler(event: Event) -> None:
    """Function can do anything about the event,
    here we're just logging it if it's a pipeline-level event.
    """
    if event.event_type.is_pipeline_event:
        logger.warning(event)


async def main(neo4j_driver: neo4j.Driver) -> PipelineResult:
    """This is where we define and run the Lexical Graph builder pipeline, instantiating
    a few components:

    - Text Splitter: to split the text into manageable chunks of fixed size
    - Chunk Embedder: to embed the chunks' text
    - Lexical Graph Builder: to build the lexical graph, ie creating the chunk nodes and relationships between them
    - KG writer: save the lexical graph to Neo4j
    """
    pipe = Pipeline(
        callback=event_handler,
    )
    # define the components
    pipe.add_component(
        FixedSizeSplitter(chunk_size=300, chunk_overlap=10, approximate=False),
        "splitter",
    )
    pipe.add_component(
        LexicalGraphBuilder(),
        "lexical_graph_builder",
    )
    pipe.add_component(Neo4jWriter(neo4j_driver), "writer")
    # define the execution order of component
    # and how the output of previous components must be used
    pipe.connect(
        "splitter", "lexical_graph_builder", input_config={"text_chunks": "splitter"}
    )
    pipe.connect(
        "lexical_graph_builder",
        "writer",
        input_config={
            "graph": "lexical_graph_builder.graph",
            "lexical_graph_config": "lexical_graph_builder.config",
        },
    )
    # user input:
    # the initial text
    # and the list of entities and relations we are looking for
    pipe_inputs = {
        "splitter": {
            "text": """Albert Einstein was a German physicist born in 1879 who
            wrote many groundbreaking papers especially about general relativity
            and quantum mechanics. He worked for many different institutions, including
            the University of Bern in Switzerland and the University of Oxford."""
        },
        "lexical_graph_builder": {
            "document_info": {
                # 'path' can be anything
                "path": "example/pipeline_with_notifications"
            },
        },
    }
    # run the pipeline
    return await pipe.run(pipe_inputs)


if __name__ == "__main__":
    with neo4j.GraphDatabase.driver(
        "bolt://localhost:7687", auth=("neo4j", "password")
    ) as driver:
        print(asyncio.run(main(driver)))



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

from typing import Any, Dict, List
from unittest.mock import MagicMock, patch

import pytest
from neo4j import Driver, Query
from neo4j_graphrag.schema import (
    BASE_ENTITY_LABEL,
    BASE_KG_BUILDER_LABEL,
    EXCLUDED_LABELS,
    EXCLUDED_RELS,
    INDEX_QUERY,
    LIST_LIMIT,
    NODE_PROPERTIES_QUERY,
    REL_PROPERTIES_QUERY,
    REL_QUERY,
    _value_sanitize,
    format_schema,
    get_enhanced_schema_cypher,
    get_schema,
    get_structured_schema,
)


def _query_return_value(*args: Any, **kwargs: Any) -> list[Any]:
    query = kwargs.get("query", args[1] if len(args) > 1 else "")
    if NODE_PROPERTIES_QUERY in query:
        return [
            {
                "output": {
                    "properties": [{"property": "property_a", "type": "STRING"}],
                    "label": "LabelA",
                }
            }
        ]
    if REL_PROPERTIES_QUERY in query:
        return [
            {
                "output": {
                    "type": "REL_TYPE",
                    "properties": [{"property": "rel_prop", "type": "STRING"}],
                }
            }
        ]
    if REL_QUERY in query:
        return [
            {"output": {"start": "LabelA", "type": "REL_TYPE", "end": "LabelB"}},
            {"output": {"start": "LabelA", "type": "REL_TYPE", "end": "LabelC"}},
        ]
    if "SHOW CONSTRAINTS" == query:
        return ["fake constraints"]
    if INDEX_QUERY == query:
        return ["fake indexes"]

    raise AssertionError("Unexpected query")


@patch("neo4j_graphrag.schema.query_database", side_effect=_query_return_value)
def test_get_schema_ensure_formatted_response(driver: Driver) -> None:
    result = get_schema(driver)
    assert (
        result
        == """Node properties:
LabelA {property_a: STRING}
Relationship properties:
REL_TYPE {rel_prop: STRING}
The relationships:
(:LabelA)-[:REL_TYPE]->(:LabelB)
(:LabelA)-[:REL_TYPE]->(:LabelC)"""
    )


def test_get_structured_schema_happy_path(driver: MagicMock) -> None:
    get_structured_schema(driver)
    assert 5 == driver.execute_query.call_count
    calls = driver.execute_query.call_args_list

    args, kwargs = calls[0]
    query_obj = args[0]
    assert isinstance(query_obj, Query)
    assert query_obj.text == NODE_PROPERTIES_QUERY
    assert query_obj.timeout is None
    assert kwargs["database_"] is None
    assert kwargs["parameters_"] == {
        "EXCLUDED_LABELS": EXCLUDED_LABELS + [BASE_ENTITY_LABEL, BASE_KG_BUILDER_LABEL],
        "SAMPLE": 1000,
    }

    args, kwargs = calls[1]
    query_obj = args[0]
    assert isinstance(query_obj, Query)
    assert query_obj.text == REL_PROPERTIES_QUERY
    assert query_obj.timeout is None
    assert kwargs["database_"] is None
    assert kwargs["parameters_"] == {"EXCLUDED_LABELS": EXCLUDED_RELS, "SAMPLE": 1000}

    args, kwargs = calls[2]
    query_obj = args[0]
    assert isinstance(query_obj, Query)
    assert query_obj.text == REL_QUERY
    assert query_obj.timeout is None
    assert kwargs["database_"] is None
    assert kwargs["parameters_"] == {
        "EXCLUDED_LABELS": EXCLUDED_LABELS + [BASE_ENTITY_LABEL, BASE_KG_BUILDER_LABEL],
        "SAMPLE": 1000,
    }

    args, kwargs = calls[3]
    query_obj = args[0]
    assert isinstance(query_obj, Query)
    assert query_obj.text == "SHOW CONSTRAINTS"
    assert query_obj.timeout is None
    assert kwargs["database_"] is None
    assert kwargs["parameters_"] == {}

    args, kwargs = calls[4]
    query_obj = args[0]
    assert isinstance(query_obj, Query)
    assert query_obj.text == INDEX_QUERY
    assert query_obj.timeout is None
    assert kwargs["database_"] is None
    assert kwargs["parameters_"] == {}


@patch("neo4j_graphrag.schema.query_database", side_effect=_query_return_value)
def test_get_schema_ensure_structured_response(driver: MagicMock) -> None:
    result = get_structured_schema(driver)

    assert result["node_props"]["LabelA"] == [
        {"property": "property_a", "type": "STRING"}
    ]
    assert result["rel_props"]["REL_TYPE"] == [
        {"property": "rel_prop", "type": "STRING"}
    ]
    assert result["relationships"] == [
        {"end": "LabelB", "start": "LabelA", "type": "REL_TYPE"},
        {"end": "LabelC", "start": "LabelA", "type": "REL_TYPE"},
    ]
    assert result["metadata"]["constraint"] == ["fake constraints"]
    assert result["metadata"]["index"] == ["fake indexes"]


@pytest.mark.parametrize(
    "description, input_value, expected_output",
    [
        (
            "Small list",
            {"key1": "value1", "small_list": list(range(15))},
            {"key1": "value1", "small_list": list(range(15))},
        ),
        (
            "Oversized list",
            {"key1": "value1", "oversized_list": list(range(LIST_LIMIT + 1))},
            {"key1": "value1"},
        ),
        (
            "Nested oversized list",
            {"key1": "value1", "oversized_list": {"key": list(range(150))}},
            {"key1": "value1", "oversized_list": {}},
        ),
        (
            "Dict in list",
            {
                "key1": "value1",
                "oversized_list": [1, 2, {"key": list(range(LIST_LIMIT + 1))}],
            },
            {"key1": "value1", "oversized_list": [1, 2, {}]},
        ),
        (
            "Dict in nested list",
            {
                "key1": "value1",
                "deeply_nested_lists": [
                    [[[{"final_nested_key": list(range(LIST_LIMIT + 1))}]]]
                ],
            },
            {"key1": "value1", "deeply_nested_lists": [[[[{}]]]]},
        ),
        (
            "Bare oversized list",
            list(range(LIST_LIMIT + 1)),
            None,
        ),
        (
            "None value",
            None,
            None,
        ),
    ],
)
def test__value_sanitize(
    description: str, input_value: Dict[str, Any], expected_output: Any
) -> None:
    """Test the _value_sanitize function."""
    assert (
        _value_sanitize(input_value) == expected_output
    ), f"Failed test case: {description}"


@pytest.mark.parametrize(
    "description, schema, is_enhanced, expected_output",
    [
        (
            "Enhanced, string property with high distinct count",
            {
                "node_props": {
                    "Person": [
                        {
                            "property": "name",
                            "type": "STRING",
                            "values": ["Alice", "Bob", "Charlie"],
                            "distinct_count": 11,
                        }
                    ]
                },
                "rel_props": {},
                "relationships": [],
            },
            True,
            (
                "Node properties:\n"
                "- **Person**\n"
                '  - `name`: STRING Example: "Alice"\n'
                "Relationship properties:\n"
                "\n"
                "The relationships:\n"
            ),
        ),
        (
            "Enhanced, string property with low distinct count",
            {
                "node_props": {
                    "Animal": [
                        {
                            "property": "species",
                            "type": "STRING",
                            "values": ["Cat", "Dog"],
                            "distinct_count": 2,
                        }
                    ]
                },
                "rel_props": {},
                "relationships": [],
            },
            True,
            (
                "Node properties:\n"
                "- **Animal**\n"
                "  - `species`: STRING Available options: ['Cat', 'Dog']\n"
                "Relationship properties:\n"
                "\n"
                "The relationships:\n"
            ),
        ),
        (
            "Enhanced, numeric property with min and max",
            {
                "node_props": {
                    "Person": [
                        {"property": "age", "type": "INTEGER", "min": 20, "max": 70}
                    ]
                },
                "rel_props": {},
                "relationships": [],
            },
            True,
            (
                "Node properties:\n"
                "- **Person**\n"
                "  - `age`: INTEGER Min: 20, Max: 70\n"
                "Relationship properties:\n"
                "\n"
                "The relationships:\n"
            ),
        ),
        (
            "Enhanced, numeric property with values",
            {
                "node_props": {
                    "Event": [
                        {
                            "property": "date",
                            "type": "DATE",
                            "values": ["2021-01-01", "2021-01-02"],
                        }
                    ]
                },
                "rel_props": {},
                "relationships": [],
            },
            True,
            (
                "Node properties:\n"
                "- **Event**\n"
                '  - `date`: DATE Example: "2021-01-01"\n'
                "Relationship properties:\n"
                "\n"
                "The relationships:\n"
            ),
        ),
        (
            "Enhanced, list property that should be skipped",
            {
                "node_props": {
                    "Document": [
                        {
                            "property": "embedding",
                            "type": "LIST",
                            "min_size": 150,
                            "max_size": 200,
                        }
                    ]
                },
                "rel_props": {},
                "relationships": [],
            },
            True,
            (
                "Node properties:\n"
                "- **Document**\n"
                "Relationship properties:\n"
                "\n"
                "The relationships:\n"
            ),
        ),
        (
            "Enhanced, list property that should be included",
            {
                "node_props": {
                    "Document": [
                        {
                            "property": "keywords",
                            "type": "LIST",
                            "min_size": 2,
                            "max_size": 5,
                        }
                    ]
                },
                "rel_props": {},
                "relationships": [],
            },
            True,
            (
                "Node properties:\n"
                "- **Document**\n"
                "  - `keywords`: LIST Min Size: 2, Max Size: 5\n"
                "Relationship properties:\n"
                "\n"
                "The relationships:\n"
            ),
        ),
        (
            "Enhanced, relationship string property with high distinct count",
            {
                "node_props": {},
                "rel_props": {
                    "KNOWS": [
                        {
                            "property": "since",
                            "type": "STRING",
                            "values": ["2000", "2001", "2002"],
                            "distinct_count": 15,
                        }
                    ]
                },
                "relationships": [],
            },
            True,
            (
                "Node properties:\n"
                "\n"
                "Relationship properties:\n"
                "- **KNOWS**\n"
                '  - `since`: STRING Example: "2000"\n'
                "The relationships:\n"
            ),
        ),
        (
            "Enhanced, relationship string property with low distinct count",
            {
                "node_props": {},
                "rel_props": {
                    "LIKES": [
                        {
                            "property": "intensity",
                            "type": "STRING",
                            "values": ["High", "Medium", "Low"],
                            "distinct_count": 3,
                        }
                    ]
                },
                "relationships": [],
            },
            True,
            (
                "Node properties:\n"
                "\n"
                "Relationship properties:\n"
                "- **LIKES**\n"
                "  - `intensity`: STRING Available options: ['High', 'Medium', 'Low']\n"
                "The relationships:\n"
            ),
        ),
        (
            "Enhanced, relationship numeric property with min and max",
            {
                "node_props": {},
                "rel_props": {
                    "WORKS_WITH": [
                        {
                            "property": "since",
                            "type": "INTEGER",
                            "min": 1995,
                            "max": 2020,
                        }
                    ]
                },
                "relationships": [],
            },
            True,
            (
                "Node properties:\n"
                "\n"
                "Relationship properties:\n"
                "- **WORKS_WITH**\n"
                "  - `since`: INTEGER Min: 1995, Max: 2020\n"
                "The relationships:\n"
            ),
        ),
        (
            "Enhanced, relationship list property that should be skipped",
            {
                "node_props": {},
                "rel_props": {
                    "KNOWS": [
                        {
                            "property": "embedding",
                            "type": "LIST",
                            "min_size": 150,
                            "max_size": 200,
                        }
                    ]
                },
                "relationships": [],
            },
            True,
            (
                "Node properties:\n"
                "\n"
                "Relationship properties:\n"
                "- **KNOWS**\n"
                "The relationships:\n"
            ),
        ),
        (
            "Enhanced, relationship list property that should be included",
            {
                "node_props": {},
                "rel_props": {
                    "KNOWS": [
                        {
                            "property": "messages",
                            "type": "LIST",
                            "min_size": 2,
                            "max_size": 5,
                        }
                    ]
                },
                "relationships": [],
            },
            True,
            (
                "Node properties:\n"
                "\n"
                "Relationship properties:\n"
                "- **KNOWS**\n"
                "  - `messages`: LIST Min Size: 2, Max Size: 5\n"
                "The relationships:\n"
            ),
        ),
        (
            "Enhanced, relationship numeric property without min and max",
            {
                "node_props": {},
                "rel_props": {
                    "OWES": [
                        {
                            "property": "amount",
                            "type": "FLOAT",
                            "values": [3.14, 2.71],
                        }
                    ]
                },
                "relationships": [],
            },
            True,
            (
                "Node properties:\n"
                "\n"
                "Relationship properties:\n"
                "- **OWES**\n"
                '  - `amount`: FLOAT Example: "3.14"\n'
                "The relationships:\n"
            ),
        ),
        (
            "Enhanced, property with empty values list",
            {
                "node_props": {
                    "Person": [
                        {
                            "property": "name",
                            "type": "STRING",
                            "values": [],
                            "distinct_count": 15,
                        }
                    ]
                },
                "rel_props": {},
                "relationships": [],
            },
            True,
            (
                "Node properties:\n"
                "- **Person**\n"
                "  - `name`: STRING \n"
                "Relationship properties:\n"
                "\n"
                "The relationships:\n"
            ),
        ),
        (
            "Enhanced, property with missing values",
            {
                "node_props": {
                    "Person": [
                        {
                            "property": "name",
                            "type": "STRING",
                            "distinct_count": 15,
                        }
                    ]
                },
                "rel_props": {},
                "relationships": [],
            },
            True,
            (
                "Node properties:\n"
                "- **Person**\n"
                "  - `name`: STRING \n"
                "Relationship properties:\n"
                "\n"
                "The relationships:\n"
            ),
        ),
    ],
)
def test_format_schema(
    description: str, schema: Dict[str, Any], is_enhanced: bool, expected_output: str
) -> None:
    result = format_schema(schema, is_enhanced)
    assert result == expected_output, f"Failed test case: {description}"


@pytest.mark.parametrize(
    "description, structured_schema, label_or_type, properties, exhaustive, sample_size, is_relationship, expected_query",
    [
        (
            "Exhaustive, string property",
            {"metadata": {"index": []}},
            "Person",
            [{"property": "name", "type": "STRING"}],
            True,
            5,
            False,
            (
                "MATCH (n:`Person`)\n"
                "WITH collect(distinct substring(toString(n.`name`), 0, 50)) AS `name_values`\n"
                "RETURN {`name`: {values: `name_values`[..10], distinct_count: size(`name_values`)}} AS output"
            ),
        ),
        (
            "Non-exhaustive, string property",
            {"metadata": {"index": []}},
            "Person",
            [{"property": "name", "type": "STRING"}],
            False,
            5,
            False,
            (
                "MATCH (n:`Person`) WITH n LIMIT 5\n"
                "WITH collect(distinct substring(toString(n.`name`), 0, 50)) AS `name_values`\n"
                "RETURN {`name`: {values: `name_values`}} AS output"
            ),
        ),
        (
            "Exhaustive, integer property",
            {"metadata": {"index": []}},
            "Person",
            [{"property": "age", "type": "INTEGER"}],
            True,
            5,
            False,
            (
                "MATCH (n:`Person`)\n"
                "WITH min(n.`age`) AS `age_min`,\n"
                "     max(n.`age`) AS `age_max`,\n"
                "     count(distinct n.`age`) AS `age_distinct`\n"
                "RETURN {`age`: {min: toString(`age_min`), max: toString(`age_max`), distinct_count: `age_distinct`}} AS output"
            ),
        ),
        (
            "Non-exhaustive, integer property",
            {"metadata": {"index": []}},
            "Person",
            [{"property": "age", "type": "INTEGER"}],
            False,
            5,
            False,
            (
                "MATCH (n:`Person`) WITH n LIMIT 5\n"
                "WITH collect(distinct toString(n.`age`)) AS `age_values`\n"
                "RETURN {`age`: {values: `age_values`}} AS output"
            ),
        ),
        (
            "Exhaustive, float property",
            {"metadata": {"index": []}},
            "Person",
            [{"property": "salary", "type": "FLOAT"}],
            True,
            5,
            False,
            (
                "MATCH (n:`Person`)\n"
                "WITH min(n.`salary`) AS `salary_min`,\n"
                "     max(n.`salary`) AS `salary_max`,\n"
                "     count(distinct n.`salary`) AS `salary_distinct`\n"
                "RETURN {`salary`: {min: toString(`salary_min`), max: toString(`salary_max`), distinct_count: `salary_distinct`}} AS output"
            ),
        ),
        (
            "Non-exhaustive, float property",
            {"metadata": {"index": []}},
            "Person",
            [{"property": "salary", "type": "FLOAT"}],
            False,
            5,
            False,
            (
                "MATCH (n:`Person`) WITH n LIMIT 5\n"
                "WITH collect(distinct toString(n.`salary`)) AS `salary_values`\n"
                "RETURN {`salary`: {values: `salary_values`}} AS output"
            ),
        ),
        (
            "Exhaustive, date property",
            {"metadata": {"index": []}},
            "Person",
            [{"property": "birth_date", "type": "DATE"}],
            True,
            5,
            False,
            (
                "MATCH (n:`Person`)\n"
                "WITH min(n.`birth_date`) AS `birth_date_min`,\n"
                "     max(n.`birth_date`) AS `birth_date_max`,\n"
                "     count(distinct n.`birth_date`) AS `birth_date_distinct`\n"
                "RETURN {`birth_date`: {min: toString(`birth_date_min`), max: toString(`birth_date_max`), distinct_count: `birth_date_distinct`}} AS output"
            ),
        ),
        (
            "Non-exhaustive, date property",
            {"metadata": {"index": []}},
            "Person",
            [{"property": "birth_date", "type": "DATE"}],
            False,
            5,
            False,
            (
                "MATCH (n:`Person`) WITH n LIMIT 5\n"
                "WITH collect(distinct toString(n.`birth_date`)) AS `birth_date_values`\n"
                "RETURN {`birth_date`: {values: `birth_date_values`}} AS output"
            ),
        ),
        (
            "Exhaustive, date time property",
            {"metadata": {"index": []}},
            "Person",
            [{"property": "birth_date", "type": "DATE_TIME"}],
            True,
            5,
            False,
            (
                "MATCH (n:`Person`)\n"
                "WITH min(n.`birth_date`) AS `birth_date_min`,\n"
                "     max(n.`birth_date`) AS `birth_date_max`,\n"
                "     count(distinct n.`birth_date`) AS `birth_date_distinct`\n"
                "RETURN {`birth_date`: {min: toString(`birth_date_min`), max: toString(`birth_date_max`), distinct_count: `birth_date_distinct`}} AS output"
            ),
        ),
        (
            "Non-exhaustive, date time property",
            {"metadata": {"index": []}},
            "Person",
            [{"property": "birth_date", "type": "DATE_TIME"}],
            False,
            5,
            False,
            (
                "MATCH (n:`Person`) WITH n LIMIT 5\n"
                "WITH collect(distinct toString(n.`birth_date`)) AS `birth_date_values`\n"
                "RETURN {`birth_date`: {values: `birth_date_values`}} AS output"
            ),
        ),
        (
            "Exhaustive, local date time property",
            {"metadata": {"index": []}},
            "Person",
            [{"property": "birth_date", "type": "LOCAL_DATE_TIME"}],
            True,
            5,
            False,
            (
                "MATCH (n:`Person`)\n"
                "WITH min(n.`birth_date`) AS `birth_date_min`,\n"
                "     max(n.`birth_date`) AS `birth_date_max`,\n"
                "     count(distinct n.`birth_date`) AS `birth_date_distinct`\n"
                "RETURN {`birth_date`: {min: toString(`birth_date_min`), max: toString(`birth_date_max`), distinct_count: `birth_date_distinct`}} AS output"
            ),
        ),
        (
            "Non-exhaustive, local date time property",
            {"metadata": {"index": []}},
            "Person",
            [{"property": "birth_date", "type": "LOCAL_DATE_TIME"}],
            False,
            5,
            False,
            (
                "MATCH (n:`Person`) WITH n LIMIT 5\n"
                "WITH collect(distinct toString(n.`birth_date`)) AS `birth_date_values`\n"
                "RETURN {`birth_date`: {values: `birth_date_values`}} AS output"
            ),
        ),
        (
            "Exhaustive, list property",
            {"metadata": {"index": []}},
            "Person",
            [{"property": "tags", "type": "LIST"}],
            True,
            5,
            False,
            (
                "MATCH (n:`Person`)\n"
                "WITH min(size(n.`tags`)) AS `tags_size_min`, max(size(n.`tags`)) AS `tags_size_max`\n"
                "RETURN {`tags`: {min_size: `tags_size_min`, max_size: `tags_size_max`}} AS output"
            ),
        ),
        (
            "Non-exhaustive, list property",
            {"metadata": {"index": []}},
            "Person",
            [{"property": "tags", "type": "LIST"}],
            False,
            5,
            False,
            (
                "MATCH (n:`Person`) WITH n LIMIT 5\n"
                "WITH min(size(n.`tags`)) AS `tags_size_min`, max(size(n.`tags`)) AS `tags_size_max`\n"
                "RETURN {`tags`: {min_size: `tags_size_min`, max_size: `tags_size_max`}} AS output"
            ),
        ),
        (
            "Exhaustive, boolean property",
            {"metadata": {"index": []}},
            "Person",
            [{"property": "active", "type": "BOOLEAN"}],
            True,
            5,
            False,
            "MATCH (n:`Person`)\nRETURN {} AS output",
        ),
        (
            "Non-exhaustive, boolean property",
            {"metadata": {"index": []}},
            "Person",
            [{"property": "active", "type": "BOOLEAN"}],
            False,
            5,
            False,
            "MATCH (n:`Person`) WITH n LIMIT 5\nRETURN {} AS output",
        ),
        (
            "Exhaustive, point property",
            {"metadata": {"index": []}},
            "Person",
            [{"property": "location", "type": "POINT"}],
            True,
            5,
            False,
            "MATCH (n:`Person`)\nRETURN {} AS output",
        ),
        (
            "Non-exhaustive, point property",
            {"metadata": {"index": []}},
            "Person",
            [{"property": "location", "type": "POINT"}],
            False,
            5,
            False,
            "MATCH (n:`Person`) WITH n LIMIT 5\nRETURN {} AS output",
        ),
        (
            "Exhaustive, duration property",
            {"metadata": {"index": []}},
            "Journey",
            [{"property": "duration", "type": "DURATION"}],
            True,
            5,
            False,
            ("MATCH (n:`Journey`)\nRETURN {} AS output"),
        ),
        (
            "Non-exhaustive, duration property",
            {"metadata": {"index": []}},
            "Journey",
            [{"property": "duration", "type": "DURATION"}],
            False,
            5,
            False,
            "MATCH (n:`Journey`) WITH n LIMIT 5\nRETURN {} AS output",
        ),
        (
            "Relationship property",
            {"metadata": {"index": []}},
            "FRIENDS_WITH",
            [{"property": "since", "type": "INTEGER"}],
            True,
            5,
            True,
            (
                "MATCH ()-[n:`FRIENDS_WITH`]->()\n"
                "WITH min(n.`since`) AS `since_min`,\n"
                "     max(n.`since`) AS `since_max`,\n"
                "     count(distinct n.`since`) AS `since_distinct`\n"
                "RETURN {`since`: {min: toString(`since_min`), max: toString(`since_max`), distinct_count: `since_distinct`}} AS output"
            ),
        ),
    ],
)
def test_get_enhanced_schema_cypher(
    driver: MagicMock,
    description: str,
    structured_schema: Dict[str, Any],
    label_or_type: str,
    properties: List[Dict[str, Any]],
    exhaustive: bool,
    sample_size: int,
    is_relationship: bool,
    expected_query: str,
) -> None:
    query = get_enhanced_schema_cypher(
        driver=driver,
        structured_schema=structured_schema,
        label_or_type=label_or_type,
        properties=properties,
        exhaustive=exhaustive,
        sample_size=sample_size,
        is_relationship=is_relationship,
    )
    assert query == expected_query, f"Failed test case: {description}"


def test_enhanced_schema_cypher_integer_exhaustive_false_with_index(
    driver: MagicMock,
) -> None:
    structured_schema = {
        "metadata": {
            "index": [
                {
                    "label": "Person",
                    "properties": ["age"],
                    "type": "RANGE",
                }
            ]
        }
    }
    properties = [{"property": "age", "type": "INTEGER"}]
    query = get_enhanced_schema_cypher(
        driver=driver,
        structured_schema=structured_schema,
        label_or_type="Person",
        properties=properties,
        exhaustive=False,
    )
    assert query == (
        "MATCH (n:`Person`) WITH n LIMIT 5\n"
        "WITH min(n.`age`) AS `age_min`,\n"
        "     max(n.`age`) AS `age_max`,\n"
        "     count(distinct n.`age`) AS `age_distinct`\n"
        "RETURN {`age`: {min: toString(`age_min`), max: toString(`age_max`), distinct_count: `age_distinct`}} AS output"
    )


@patch("neo4j_graphrag.schema.query_database")
def test_enhanced_schema_cypher_string_exhaustive_false_with_index(
    query_database_mock: MagicMock,
    driver: MagicMock,
) -> None:
    structured_schema = {
        "metadata": {
            "index": [
                {
                    "label": "Person",
                    "properties": ["status"],
                    "type": "RANGE",
                    "size": 5,
                    "distinctValues": 5,
                }
            ]
        }
    }
    query_database_mock.return_value = [{"value": ["Single", "Married", "Divorced"]}]
    properties = [{"property": "status", "type": "STRING"}]
    query = get_enhanced_schema_cypher(
        driver=driver,
        structured_schema=structured_schema,
        label_or_type="Person",
        properties=properties,
        exhaustive=False,
    )
    assert query == (
        "MATCH (n:`Person`) WITH n LIMIT 5\n\n"
        "RETURN {`status`: {values: ['Single', 'Married', 'Divorced'], distinct_count: 3}} AS output"
    )



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from unittest.mock import MagicMock

import pytest
from neo4j_graphrag.experimental.components.embedder import TextChunkEmbedder
from neo4j_graphrag.experimental.components.types import (
    TextChunk,
    TextChunks,
)


@pytest.mark.asyncio
async def test_text_chunk_embedder_run(embedder: MagicMock) -> None:
    embedder.async_embed_query.return_value = [1.0, 2.0, 3.0]
    text_chunk_embedder = TextChunkEmbedder(embedder=embedder)
    text_chunks = TextChunks(
        chunks=[TextChunk(text="may thy knife chip and shatter", index=0)]
    )
    embedded_chunks = await text_chunk_embedder.run(text_chunks)
    embedder.async_embed_query.assert_called_once_with("may thy knife chip and shatter")
    assert isinstance(embedded_chunks, TextChunks)
    for chunk in embedded_chunks.chunks:
        assert isinstance(chunk, TextChunk)
        assert chunk.metadata is not None
        assert "embedding" in chunk.metadata.keys()
        assert isinstance(chunk.metadata["embedding"], list)
        for i in chunk.metadata["embedding"]:
            assert isinstance(i, float)



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

import json
from unittest.mock import MagicMock

import pytest
from neo4j_graphrag.exceptions import LLMGenerationError
from neo4j_graphrag.experimental.components.entity_relation_extractor import (
    LLMEntityRelationExtractor,
    OnError,
    balance_curly_braces,
    fix_invalid_json,
)
from neo4j_graphrag.experimental.components.types import (
    DocumentInfo,
    Neo4jGraph,
    TextChunk,
    TextChunks,
)
from neo4j_graphrag.experimental.pipeline.exceptions import InvalidJSONError
from neo4j_graphrag.llm import LLMInterface, LLMResponse
from neo4j_graphrag.llm import AnthropicLLM, OpenAILLM, VertexAILLM
from unittest.mock import patch


@pytest.mark.asyncio
async def test_extractor_happy_path_no_entities_no_document() -> None:
    llm = MagicMock(spec=LLMInterface)
    llm.ainvoke.return_value = LLMResponse(content='{"nodes": [], "relationships": []}')

    extractor = LLMEntityRelationExtractor(
        llm=llm,
    )
    chunks = TextChunks(chunks=[TextChunk(text="some text", index=0)])
    result = await extractor.run(chunks=chunks)
    assert isinstance(result, Neo4jGraph)
    # only one Chunk node (no document info provided)
    assert len(result.nodes) == 1
    assert result.nodes[0].label == "Chunk"
    assert result.relationships == []


@pytest.mark.asyncio
async def test_extractor_happy_path_no_entities() -> None:
    llm = MagicMock(spec=LLMInterface)
    llm.ainvoke.return_value = LLMResponse(content='{"nodes": [], "relationships": []}')

    extractor = LLMEntityRelationExtractor(
        llm=llm,
    )
    chunks = TextChunks(chunks=[TextChunk(text="some text", index=0)])
    document_info = DocumentInfo(path="path")
    result = await extractor.run(chunks=chunks, document_info=document_info)
    assert isinstance(result, Neo4jGraph)
    # one Chunk node and one Document node
    assert len(result.nodes) == 2
    assert set(n.label for n in result.nodes) == {"Chunk", "Document"}
    assert len(result.relationships) == 1
    assert result.relationships[0].type == "FROM_DOCUMENT"


@pytest.mark.asyncio
async def test_extractor_happy_path_no_entities_no_lexical_graph() -> None:
    llm = MagicMock(spec=LLMInterface)
    llm.ainvoke.return_value = LLMResponse(content='{"nodes": [], "relationships": []}')

    extractor = LLMEntityRelationExtractor(
        llm=llm,
        create_lexical_graph=False,
    )
    chunks = TextChunks(chunks=[TextChunk(text="some text", index=0)])
    document_info = DocumentInfo(path="path")
    graph = await extractor.run(chunks=chunks, document_info=document_info)
    assert graph.nodes == []
    assert graph.relationships == []


@pytest.mark.asyncio
async def test_extractor_happy_path_non_empty_result() -> None:
    llm = MagicMock(spec=LLMInterface)
    llm.ainvoke.return_value = LLMResponse(
        content='{"nodes": [{"id": "0", "label": "Person", "properties": {}}], "relationships": []}'
    )

    extractor = LLMEntityRelationExtractor(
        llm=llm,
    )
    chunks = TextChunks(chunks=[TextChunk(text="some text", index=0)])
    document_info = DocumentInfo(path="path")
    result = await extractor.run(chunks=chunks, document_info=document_info)
    assert isinstance(result, Neo4jGraph)
    assert len(result.nodes) == 3
    doc = result.nodes[0]
    assert doc.label == "Document"
    chunk_entity = result.nodes[1]
    assert chunk_entity.label == "Chunk"
    entity = result.nodes[2]
    assert entity.id == f"{chunk_entity.id}:0"
    assert entity.label == "Person"
    assert len(result.relationships) == 2
    assert result.relationships[0].type == "FROM_DOCUMENT"
    assert result.relationships[0].start_node_id == f"{chunk_entity.id}"
    assert result.relationships[0].end_node_id == f"{doc.id}"
    assert result.relationships[1].type == "FROM_CHUNK"


@pytest.mark.asyncio
async def test_extractor_missing_entity_id() -> None:
    llm = MagicMock(spec=LLMInterface)
    llm.ainvoke.return_value = LLMResponse(
        content='{"nodes": [{"label": "Person", "properties": {}}], "relationships": []}'
    )
    extractor = LLMEntityRelationExtractor(
        llm=llm,
    )
    chunks = TextChunks(chunks=[TextChunk(text="some text", index=0)])
    with pytest.raises(LLMGenerationError):
        await extractor.run(chunks=chunks)


@pytest.mark.asyncio
async def test_extractor_llm_ainvoke_failed() -> None:
    llm = MagicMock(spec=LLMInterface)
    llm.ainvoke.side_effect = LLMGenerationError()

    extractor = LLMEntityRelationExtractor(
        llm=llm,
    )
    chunks = TextChunks(chunks=[TextChunk(text="some text", index=0)])
    with pytest.raises(LLMGenerationError):
        await extractor.run(chunks=chunks)


@pytest.mark.asyncio
async def test_extractor_llm_unfixable_json() -> None:
    llm = MagicMock(spec=LLMInterface)
    llm.ainvoke.return_value = LLMResponse(
        content='{"nodes": [{"id": "0", "label": "Person", "properties": {}}], "relationships": }'
    )

    extractor = LLMEntityRelationExtractor(
        llm=llm,
    )
    chunks = TextChunks(chunks=[TextChunk(text="some text", index=0)])

    with pytest.raises(LLMGenerationError):
        await extractor.run(chunks=chunks)


@pytest.mark.asyncio
async def test_extractor_llm_invalid_json() -> None:
    """Test what happens when the returned JSON is valid JSON but
    does not match the expected Pydantic model"""
    llm = MagicMock(spec=LLMInterface)
    llm.ainvoke.return_value = LLMResponse(
        # missing "label" for entity
        content='{"nodes": [{"id": 0, "entity_type": "Person", "properties": {}}], "relationships": []}'
    )

    extractor = LLMEntityRelationExtractor(
        llm=llm,
    )
    chunks = TextChunks(chunks=[TextChunk(text="some text", index=0)])
    with pytest.raises(LLMGenerationError):
        await extractor.run(chunks=chunks)


@pytest.mark.asyncio
async def test_extractor_llm_invalid_json_is_a_list() -> None:
    """Test what happens when the returned JSON is a valid JSON list,
    but it does not match the expected Pydantic model"""
    llm = MagicMock(spec=LLMInterface)
    llm.ainvoke.return_value = LLMResponse(
        # missing "label" for entity
        content='[{"nodes": [{"id": 0, "entity_type": "Person", "properties": {}}], "relationships": []}]'
    )

    extractor = LLMEntityRelationExtractor(
        llm=llm,
    )
    chunks = TextChunks(chunks=[TextChunk(text="some text", index=0)])
    with pytest.raises(LLMGenerationError):
        await extractor.run(chunks=chunks)


@pytest.mark.asyncio
async def test_extractor_llm_badly_formatted_json_gets_fixed() -> None:
    llm = MagicMock(spec=LLMInterface)
    llm.ainvoke.return_value = LLMResponse(
        content='{"nodes": [{"id": "0", "label": "Person", "properties": {}}], "relationships": [}'
    )

    extractor = LLMEntityRelationExtractor(
        llm=llm,
        on_error=OnError.IGNORE,
        create_lexical_graph=False,
    )
    chunks = TextChunks(chunks=[TextChunk(text="some text", index=0)])
    res = await extractor.run(chunks=chunks)

    assert len(res.nodes) == 1
    assert res.nodes[0].label == "Person"
    assert res.nodes[0].embedding_properties == {}
    assert res.relationships == []


@pytest.mark.asyncio
async def test_extractor_custom_prompt() -> None:
    llm = MagicMock(spec=LLMInterface)
    llm.ainvoke.return_value = LLMResponse(content='{"nodes": [], "relationships": []}')

    extractor = LLMEntityRelationExtractor(llm=llm, prompt_template="this is my prompt")
    chunks = TextChunks(chunks=[TextChunk(text="some text", index=0)])
    await extractor.run(chunks=chunks)
    llm.ainvoke.assert_called_once_with("this is my prompt")


def test_fix_invalid_json_empty_result() -> None:
    json_string = "invalid json"

    with patch("json_repair.repair_json", return_value=""):
        with pytest.raises(InvalidJSONError):
            fix_invalid_json(json_string)


def test_fix_unquoted_keys() -> None:
    json_string = '{name: "John", age: "30"}'
    expected_result = '{"name": "John", "age": "30"}'

    fixed_json = fix_invalid_json(json_string)

    assert json.loads(fixed_json)
    assert fixed_json == expected_result


def test_fix_unquoted_string_values() -> None:
    json_string = '{"name": John, "age": 30}'
    expected_result = '{"name": "John", "age": 30}'

    fixed_json = fix_invalid_json(json_string)

    assert json.loads(fixed_json)
    assert fixed_json == expected_result


def test_remove_trailing_commas() -> None:
    json_string = '{"name": "John", "age": 30,}'
    expected_result = '{"name": "John", "age": 30}'

    fixed_json = fix_invalid_json(json_string)

    assert json.loads(fixed_json)
    assert fixed_json == expected_result


def test_fix_excessive_braces() -> None:
    json_string = '{{"name": "John"}}'
    expected_result = '{"name": "John"}'

    fixed_json = fix_invalid_json(json_string)

    assert json.loads(fixed_json)
    assert fixed_json == expected_result


def test_fix_multiple_issues() -> None:
    json_string = '{name: John, "hobbies": ["reading", "swimming",], "age": 30}'
    expected_result = '{"name": "John", "hobbies": ["reading", "swimming"], "age": 30}'

    fixed_json = fix_invalid_json(json_string)

    assert json.loads(fixed_json)
    assert fixed_json == expected_result


def test_fix_null_values() -> None:
    json_string = '{"name": John, "nickname": null}'
    expected_result = '{"name": "John", "nickname": null}'

    fixed_json = fix_invalid_json(json_string)

    assert json.loads(fixed_json)
    assert fixed_json == expected_result


def test_fix_numeric_values() -> None:
    json_string = '{"age": 30, "score": 95.5}'
    expected_result = '{"age": 30, "score": 95.5}'

    fixed_json = fix_invalid_json(json_string)

    assert json.loads(fixed_json)
    assert fixed_json == expected_result


def test_balance_curly_braces_missing_closing() -> None:
    json_string = '{"name": "John", "hobbies": {"reading": "yes"'
    expected_result = '{"name": "John", "hobbies": {"reading": "yes"}}'

    fixed_json = balance_curly_braces(json_string)

    assert json.loads(fixed_json)
    assert fixed_json == expected_result


def test_balance_curly_braces_extra_closing() -> None:
    json_string = '{"name": "John", "hobbies": {"reading": "yes"}}}'
    expected_result = '{"name": "John", "hobbies": {"reading": "yes"}}'

    fixed_json = balance_curly_braces(json_string)

    assert json.loads(fixed_json)
    assert fixed_json == expected_result


def test_balance_curly_braces_balanced_input() -> None:
    json_string = '{"name": "John", "hobbies": {"reading": "yes"}, "age": 30}'
    expected_result = json_string

    fixed_json = balance_curly_braces(json_string)

    assert json.loads(fixed_json)
    assert fixed_json == expected_result


def test_balance_curly_braces_nested_structure() -> None:
    json_string = '{"person": {"name": "John", "hobbies": {"reading": "yes"}}}'
    expected_result = json_string

    fixed_json = balance_curly_braces(json_string)

    assert json.loads(fixed_json)
    assert fixed_json == expected_result


def test_balance_curly_braces_unbalanced_nested() -> None:
    json_string = '{"person": {"name": "John", "hobbies": {"reading": "yes"}}'
    expected_result = '{"person": {"name": "John", "hobbies": {"reading": "yes"}}}'

    fixed_json = balance_curly_braces(json_string)

    assert json.loads(fixed_json)
    assert fixed_json == expected_result


def test_balance_curly_braces_unmatched_openings() -> None:
    json_string = '{"name": "John", "hobbies": {"reading": "yes"'
    expected_result = '{"name": "John", "hobbies": {"reading": "yes"}}'

    fixed_json = balance_curly_braces(json_string)

    assert json.loads(fixed_json)
    assert fixed_json == expected_result


def test_balance_curly_braces_unmatched_closings() -> None:
    json_string = '{"name": "John", "hobbies": {"reading": "yes"}}}'
    expected_result = '{"name": "John", "hobbies": {"reading": "yes"}}'

    fixed_json = balance_curly_braces(json_string)

    assert json.loads(fixed_json)
    assert fixed_json == expected_result


def test_balance_curly_braces_complex_structure() -> None:
    json_string = (
        '{"name": "John", "details": {"age": 30, "hobbies": {"reading": "yes"}}}'
    )
    expected_result = json_string

    fixed_json = balance_curly_braces(json_string)

    assert json.loads(fixed_json)
    assert fixed_json == expected_result


def test_balance_curly_braces_incorrect_nested_closings() -> None:
    json_string = '{"key1": {"key2": {"reading": "yes"}}, "key3": {"age": 30}}}'
    expected_result = '{"key1": {"key2": {"reading": "yes"}}, "key3": {"age": 30}}'

    fixed_json = balance_curly_braces(json_string)

    assert json.loads(fixed_json)
    assert fixed_json == expected_result


def test_balance_curly_braces_braces_inside_string() -> None:
    json_string = '{"name": "John", "example": "a{b}c", "age": 30}'
    expected_result = json_string

    fixed_json = balance_curly_braces(json_string)

    assert json.loads(fixed_json)
    assert fixed_json == expected_result


def test_balance_curly_braces_unbalanced_with_string() -> None:
    json_string = '{"name": "John", "example": "a{b}c", "hobbies": {"reading": "yes"'
    expected_result = (
        '{"name": "John", "example": "a{b}c", "hobbies": {"reading": "yes"}}'
    )

    fixed_json = balance_curly_braces(json_string)

    assert json.loads(fixed_json)
    assert fixed_json == expected_result


@pytest.mark.asyncio
async def test_extractor_structured_output_with_openai() -> None:
    """Test that use_structured_output=True works with OpenAILLM."""
    llm = MagicMock(spec=OpenAILLM)
    llm.ainvoke.return_value = LLMResponse(content='{"nodes": [], "relationships": []}')

    extractor = LLMEntityRelationExtractor(
        llm=llm,
        use_structured_output=True,
    )
    chunks = TextChunks(chunks=[TextChunk(text="some text", index=0)])
    await extractor.run(chunks=chunks)

    # Verify ainvoke was called with response_format=Neo4jGraph
    llm.ainvoke.assert_called_once()
    call_args = llm.ainvoke.call_args
    assert call_args[1]["response_format"] == Neo4jGraph


@pytest.mark.asyncio
async def test_extractor_structured_output_with_vertexai() -> None:
    """Test that use_structured_output=True works with VertexAILLM."""
    llm = MagicMock(spec=VertexAILLM)
    llm.ainvoke.return_value = LLMResponse(content='{"nodes": [], "relationships": []}')

    extractor = LLMEntityRelationExtractor(
        llm=llm,
        use_structured_output=True,
    )
    chunks = TextChunks(chunks=[TextChunk(text="some text", index=0)])
    await extractor.run(chunks=chunks)

    # Verify ainvoke was called with response_format=Neo4jGraph
    llm.ainvoke.assert_called_once()
    call_args = llm.ainvoke.call_args
    assert call_args[1]["response_format"] == Neo4jGraph


def test_extractor_structured_output_unsupported_llm() -> None:
    """Test that use_structured_output=True raises error with unsupported LLMs."""
    llm = AnthropicLLM(api_key="test", model_name="claude-3-opus")

    with pytest.raises(ValueError) as exc_info:
        LLMEntityRelationExtractor(
            llm=llm,
            use_structured_output=True,
        )

    assert "Structured output is not supported" in str(exc_info.value)


@pytest.mark.asyncio
async def test_extractor_structured_output_false_uses_v1() -> None:
    """Test that use_structured_output=False uses V1 interface (prompt-based)."""
    llm = MagicMock(spec=LLMInterface)
    llm.ainvoke.return_value = LLMResponse(content='{"nodes": [], "relationships": []}')

    extractor = LLMEntityRelationExtractor(
        llm=llm,
        use_structured_output=False,  # Default behavior
    )
    chunks = TextChunks(chunks=[TextChunk(text="some text", index=0)])
    await extractor.run(chunks=chunks)

    # Verify ainvoke was called with just a string prompt (V1), not response_format
    llm.ainvoke.assert_called_once()
    call_args = llm.ainvoke.call_args
    assert isinstance(call_args[0][0], str)  # First arg is string prompt
    assert "response_format" not in call_args[1]  # No response_format kwarg



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Deprecated module: Rate limiting functionality has been moved to neo4j_graphrag.utils.rate_limit.

This module provides backward compatibility with deprecation warnings.
All new code should import from neo4j_graphrag.utils.rate_limit instead.
"""

import warnings
from typing import Any

# Import the actual implementations from the new location
from neo4j_graphrag.utils.rate_limit import (
    RateLimitHandler as _RateLimitHandler,
    NoOpRateLimitHandler as _NoOpRateLimitHandler,
    RetryRateLimitHandler as _RetryRateLimitHandler,
    rate_limit_handler as _rate_limit_handler,
    async_rate_limit_handler as _async_rate_limit_handler,
    is_rate_limit_error as _is_rate_limit_error,
    convert_to_rate_limit_error as _convert_to_rate_limit_error,
    DEFAULT_RATE_LIMIT_HANDLER as _DEFAULT_RATE_LIMIT_HANDLER,
)


def __getattr__(name: str) -> Any:
    """Handle deprecated imports with warnings."""
    deprecated_items = {
        "RateLimitHandler": _RateLimitHandler,
        "NoOpRateLimitHandler": _NoOpRateLimitHandler,
        "RetryRateLimitHandler": _RetryRateLimitHandler,
        "rate_limit_handler": _rate_limit_handler,
        "async_rate_limit_handler": _async_rate_limit_handler,
        "is_rate_limit_error": _is_rate_limit_error,
        "convert_to_rate_limit_error": _convert_to_rate_limit_error,
        "DEFAULT_RATE_LIMIT_HANDLER": _DEFAULT_RATE_LIMIT_HANDLER,
    }

    if name in deprecated_items:
        warnings.warn(
            f"{name} has been moved to neo4j_graphrag.utils.rate_limit. "
            f"Please update your imports to use 'from neo4j_graphrag.utils.rate_limit import {name}'.",
            DeprecationWarning,
            stacklevel=2,
        )
        return deprecated_items[name]

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


# Issue a single deprecation warning when the module is imported
warnings.warn(
    "The neo4j_graphrag.llm.rate_limit module has been moved to neo4j_graphrag.utils.rate_limit. "
    "Please update your imports to use 'from neo4j_graphrag.utils.rate_limit import ...'.",
    DeprecationWarning,
    stacklevel=2,
)



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from typing import Any, ClassVar, Literal, Optional, Union

import neo4j
from pydantic import field_validator

from neo4j_graphrag.embeddings import Embedder
from neo4j_graphrag.experimental.pipeline.config.base import AbstractConfig
from neo4j_graphrag.experimental.pipeline.config.object_config import (
    ComponentType,
    EmbedderType,
    LLMType,
    Neo4jDriverType,
)
from neo4j_graphrag.experimental.pipeline.config.param_resolver import (
    ParamConfig,
)
from neo4j_graphrag.experimental.pipeline.config.types import PipelineType
from neo4j_graphrag.experimental.pipeline.types.definitions import (
    ComponentDefinition,
    ConnectionDefinition,
    PipelineDefinition,
)
from neo4j_graphrag.llm import LLMInterface

logger = logging.getLogger(__name__)


class AbstractPipelineConfig(AbstractConfig):
    """This class defines the fields possibly used by all pipelines: neo4j drivers, LLMs...
    neo4j_config, llm_config can be provided by user as a single item or a dict of items.
    Validators deal with type conversion so that the field in all instances is a dict of items.
    """

    neo4j_config: dict[str, Neo4jDriverType] = {}
    llm_config: dict[str, LLMType] = {}
    embedder_config: dict[str, EmbedderType] = {}
    # extra parameters values that can be used in different places of the config file
    extras: dict[str, ParamConfig] = {}

    DEFAULT_NAME: ClassVar[str] = "default"
    """Name of the default item in dict
    """

    @field_validator("neo4j_config", mode="before")
    @classmethod
    def validate_drivers(
        cls, drivers: Union[Neo4jDriverType, dict[str, Any]]
    ) -> dict[str, Any]:
        if not isinstance(drivers, dict) or "params_" in drivers:
            return {cls.DEFAULT_NAME: drivers}
        return drivers

    @field_validator("llm_config", mode="before")
    @classmethod
    def validate_llms(cls, llms: Union[LLMType, dict[str, Any]]) -> dict[str, Any]:
        if not isinstance(llms, dict) or "class_" in llms:
            return {cls.DEFAULT_NAME: llms}
        return llms

    @field_validator("embedder_config", mode="before")
    @classmethod
    def validate_embedders(
        cls, embedders: Union[EmbedderType, dict[str, Any]]
    ) -> dict[str, Any]:
        if not isinstance(embedders, dict) or "class_" in embedders:
            return {cls.DEFAULT_NAME: embedders}
        return embedders

    def _resolve_component_definition(
        self, name: str, config: ComponentType
    ) -> ComponentDefinition:
        component = config.parse(self._global_data)
        if hasattr(config.root, "run_params_"):
            component_run_params = self.resolve_params(config.root.run_params_)
        else:
            component_run_params = {}
        component_def = ComponentDefinition(
            name=name,
            component=component,
            run_params=component_run_params,
        )
        logger.debug(f"PIPELINE_CONFIG: resolved component {component_def}")
        return component_def

    def _parse_global_data(self) -> dict[str, Any]:
        """Global data contains data that can be referenced in other parts of the
        config.

        Typically, neo4j drivers, LLMs and embedders can be referenced in component
         input parameters.
        """
        # 'extras' parameters can be referenced in other configs,
        # that's why they are parsed before the others
        # e.g., an API key used for both LLM and Embedder can be stored only
        # once in extras.
        extra_data = {
            "extras": self.resolve_params(self.extras),
        }
        logger.debug(f"PIPELINE_CONFIG: resolved 'extras': {extra_data}")
        drivers: dict[str, neo4j.Driver] = {
            driver_name: driver_config.parse(extra_data)
            for driver_name, driver_config in self.neo4j_config.items()
        }
        llms: dict[str, LLMInterface] = {
            llm_name: llm_config.parse(extra_data)
            for llm_name, llm_config in self.llm_config.items()
        }
        embedders: dict[str, Embedder] = {
            embedder_name: embedder_config.parse(extra_data)
            for embedder_name, embedder_config in self.embedder_config.items()
        }
        global_data = {
            **extra_data,
            "neo4j_config": drivers,
            "llm_config": llms,
            "embedder_config": embedders,
        }
        logger.debug(f"PIPELINE_CONFIG: resolved globals: {global_data}")
        return global_data

    def _get_components(self) -> list[ComponentDefinition]:
        return []

    def _get_connections(self) -> list[ConnectionDefinition]:
        return []

    def parse(
        self, resolved_data: Optional[dict[str, Any]] = None
    ) -> PipelineDefinition:
        """Parse the full config and returns a PipelineDefinition object, containing instantiated
        components and a list of connections.
        """
        self._global_data = self._parse_global_data()
        return PipelineDefinition(
            components=self._get_components(),
            connections=self._get_connections(),
        )

    def get_run_params(self, user_input: dict[str, Any]) -> dict[str, Any]:
        return user_input

    async def close(self) -> None:
        drivers = self._global_data.get("neo4j_config", {})
        for driver_name in drivers:
            driver = drivers[driver_name]
            logger.debug(f"PIPELINE_CONFIG: closing driver {driver_name}: {driver}")
            driver.close()

    def get_neo4j_driver_by_name(self, name: str) -> neo4j.Driver:
        drivers: dict[str, neo4j.Driver] = self._global_data.get("neo4j_config", {})
        return drivers[name]

    def get_default_neo4j_driver(self) -> neo4j.Driver:
        return self.get_neo4j_driver_by_name(self.DEFAULT_NAME)

    def get_llm_by_name(self, name: str) -> LLMInterface:
        llms: dict[str, LLMInterface] = self._global_data.get("llm_config", {})
        return llms[name]

    def get_default_llm(self) -> LLMInterface:
        return self.get_llm_by_name(self.DEFAULT_NAME)

    def get_embedder_by_name(self, name: str) -> Embedder:
        embedders: dict[str, Embedder] = self._global_data.get("embedder_config", {})
        return embedders[name]

    def get_default_embedder(self) -> Embedder:
        return self.get_embedder_by_name(self.DEFAULT_NAME)


class PipelineConfig(AbstractPipelineConfig):
    """Configuration class for raw pipelines.
    Config must contain all components and connections."""

    component_config: dict[str, ComponentType]
    connection_config: list[ConnectionDefinition]
    template_: Literal[PipelineType.NONE] = PipelineType.NONE

    def _get_connections(self) -> list[ConnectionDefinition]:
        return self.connection_config

    def _get_components(self) -> list[ComponentDefinition]:
        return [
            self._resolve_component_definition(name, component_config)
            for name, component_config in self.component_config.items()
        ]



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
"""
Example demonstrating how to create multiple domain-specific tools from retrievers.

This example shows:
1. How to create multiple tools from the same retriever type for different use cases
2. How to provide custom parameter descriptions for each tool
3. How type inference works automatically while descriptions are explicit
"""

import neo4j
from typing import cast, Any, Optional
from unittest.mock import MagicMock

from neo4j_graphrag.retrievers.base import Retriever
from neo4j_graphrag.types import RawSearchResult


class MockVectorRetriever(Retriever):
    """A mock vector retriever for demonstration purposes."""

    VERIFY_NEO4J_VERSION = False

    def __init__(self, driver: neo4j.Driver, index_name: str):
        super().__init__(driver)
        self.index_name = index_name

    def get_search_results(
        self,
        query_vector: Optional[list[float]] = None,
        query_text: Optional[str] = None,
        top_k: int = 5,
        effective_search_ratio: int = 1,
        filters: Optional[dict[str, Any]] = None,
    ) -> RawSearchResult:
        """Get vector search results (mocked for demonstration)."""
        # Return empty results for demo
        return RawSearchResult(records=[], metadata={"index": self.index_name})


def main() -> None:
    """Demonstrate creating multiple domain-specific tools from retrievers."""

    # Create mock driver (in real usage, this would be actual Neo4j driver)
    driver = cast(Any, MagicMock())

    # Create retrievers for different domains using the same retriever type
    # In practice, these would point to different vector indexes

    # Movie recommendations retriever
    movie_retriever = MockVectorRetriever(driver=driver, index_name="movie_embeddings")

    # Product search retriever
    product_retriever = MockVectorRetriever(
        driver=driver, index_name="product_embeddings"
    )

    # Document search retriever
    document_retriever = MockVectorRetriever(
        driver=driver, index_name="document_embeddings"
    )

    # Convert each retriever to a domain-specific tool with custom descriptions

    # 1. Movie recommendation tool
    movie_tool = movie_retriever.convert_to_tool(
        name="movie_search",
        description="Find movie recommendations based on plot, genre, or actor preferences",
        parameter_descriptions={
            "query_text": "Movie title, plot description, genre, or actor name",
            "query_vector": "Pre-computed embedding vector for movie search",
            "top_k": "Number of movie recommendations to return (1-20)",
            "filters": "Optional filters for genre, year, rating, etc.",
            "effective_search_ratio": "Search pool multiplier for better accuracy",
        },
    )

    # 2. Product search tool
    product_tool = product_retriever.convert_to_tool(
        name="product_search",
        description="Search for products matching customer needs and preferences",
        parameter_descriptions={
            "query_text": "Product name, description, or customer need",
            "query_vector": "Pre-computed embedding for product matching",
            "top_k": "Maximum number of product results (1-50)",
            "filters": "Filters for price range, brand, category, availability",
            "effective_search_ratio": "Breadth vs precision trade-off for search",
        },
    )

    # 3. Document search tool
    document_tool = document_retriever.convert_to_tool(
        name="document_search",
        description="Find relevant documents and knowledge articles",
        parameter_descriptions={
            "query_text": "Question, keywords, or topic to search for",
            "query_vector": "Semantic embedding for document retrieval",
            "top_k": "Number of relevant documents to retrieve (1-10)",
            "filters": "Document type, date range, or department filters",
        },
    )

    # Demonstrate that each tool has distinct, meaningful descriptions
    tools = [movie_tool, product_tool, document_tool]

    for tool in tools:
        print(f"\n=== {tool.get_name().upper()} ===")
        print(f"Description: {tool.get_description()}")
        print("Parameters:")

        params = tool.get_parameters()
        for param_name, param_def in params["properties"].items():
            required = (
                "required" if param_name in params.get("required", []) else "optional"
            )
            print(
                f"  - {param_name} ({param_def['type']}, {required}): {param_def['description']}"
            )

    # Show how the same parameter type gets different contextual descriptions
    print("\n=== PARAMETER COMPARISON ===")
    print("Same parameter 'query_text' with different contextual descriptions:")

    for tool in tools:
        params = tool.get_parameters()
        query_text_desc = params["properties"]["query_text"]["description"]
        print(f"  {tool.get_name()}: {query_text_desc}")

    print("\nSame parameter 'top_k' with different contextual descriptions:")
    for tool in tools:
        params = tool.get_parameters()
        top_k_desc = params["properties"]["top_k"]["description"]
        print(f"  {tool.get_name()}: {top_k_desc}")


if __name__ == "__main__":
    main()



import asyncio

import neo4j
from neo4j_graphrag.experimental.components.neo4j_reader import Neo4jChunkReader
from neo4j_graphrag.experimental.components.types import LexicalGraphConfig, TextChunks


async def main(driver: neo4j.Driver) -> TextChunks:
    config = LexicalGraphConfig(  # only needed to overwrite the default values
        chunk_node_label="TextPart",
    )
    reader = Neo4jChunkReader(driver)
    result = await reader.run(lexical_graph_config=config)
    return result


if __name__ == "__main__":
    with neo4j.GraphDatabase.driver(
        "bolt://localhost:7687", auth=("neo4j", "password")
    ) as driver:
        print(asyncio.run(main(driver)))



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

from unittest.mock import MagicMock

import neo4j.exceptions
import pytest
from neo4j_graphrag.exceptions import Neo4jIndexError, Neo4jInsertionError
from neo4j_graphrag.indexes import (
    create_fulltext_index,
    create_vector_index,
    drop_index_if_exists,
    upsert_vector,
    upsert_vector_on_relationship,
    upsert_vectors,
)


def test_create_vector_index_happy_path(driver: MagicMock) -> None:
    create_query = (
        "CREATE VECTOR INDEX $name IF NOT EXISTS FOR (n:People) ON n.name OPTIONS "
        "{ indexConfig: { `vector.dimensions`: toInteger($dimensions), `vector.similarity_function`: $similarity_fn } }"
    )

    create_vector_index(driver, "my-index", "People", "name", 2048, "cosine")

    driver.execute_query.assert_called_once_with(
        create_query,
        {"name": "my-index", "dimensions": 2048, "similarity_fn": "cosine"},
        database_=None,
    )


def test_create_vector_index_fail_if_exists(driver: MagicMock) -> None:
    create_query = (
        "CREATE VECTOR INDEX $name  FOR (n:People) ON n.name OPTIONS "
        "{ indexConfig: { `vector.dimensions`: toInteger($dimensions), `vector.similarity_function`: $similarity_fn } }"
    )

    create_vector_index(
        driver, "my-index", "People", "name", 2048, "cosine", fail_if_exists=True
    )

    driver.execute_query.assert_called_once_with(
        create_query,
        {"name": "my-index", "dimensions": 2048, "similarity_fn": "cosine"},
        database_=None,
    )


def test_create_vector_index_ensure_escaping(driver: MagicMock) -> None:
    create_query = (
        "CREATE VECTOR INDEX $name IF NOT EXISTS FOR (n:People) ON n.name OPTIONS "
        "{ indexConfig: { `vector.dimensions`: toInteger($dimensions), `vector.similarity_function`: $similarity_fn } }"
    )

    create_vector_index(
        driver, "my-complicated-`-index", "People", "name", 2048, "cosine"
    )

    driver.execute_query.assert_called_once_with(
        create_query,
        {
            "name": "my-complicated-`-index",
            "dimensions": 2048,
            "similarity_fn": "cosine",
        },
        database_=None,
    )


def test_create_vector_index_negative_dimension(driver: MagicMock) -> None:
    with pytest.raises(Neo4jIndexError) as excinfo:
        create_vector_index(driver, "my-index", "People", "name", -5, "cosine")
    assert "Error for inputs to create_vector_index" in str(excinfo)


def test_create_vector_index_validation_error_dimensions(driver: MagicMock) -> None:
    with pytest.raises(Neo4jIndexError) as excinfo:
        create_vector_index(driver, "my-index", "People", "name", "no-dim", "cosine")  # type: ignore
    assert "Error for inputs to create_vector_index" in str(excinfo)


def test_create_vector_index_raises_error_with_neo4j_client_error(
    driver: MagicMock,
) -> None:
    driver.execute_query.side_effect = neo4j.exceptions.ClientError
    with pytest.raises(Neo4jIndexError):
        create_vector_index(driver, "my-index", "People", "name", 2048, "cosine")


def test_create_vector_index_validation_error_similarity_fn(driver: MagicMock) -> None:
    with pytest.raises(Neo4jIndexError) as excinfo:
        create_vector_index(driver, "my-index", "People", "name", 1536, "algebra")  # type: ignore
    assert "Error for inputs to create_vector_index" in str(excinfo)


def test_drop_index_if_exists(driver: MagicMock) -> None:
    drop_query = "DROP INDEX $name IF EXISTS"

    drop_index_if_exists(driver, "my-index")

    driver.execute_query.assert_called_once_with(
        drop_query,
        {"name": "my-index"},
        database_=None,
    )


def test_drop_index_if_exists_raises_error_with_neo4j_client_error(
    driver: MagicMock,
) -> None:
    drop_query = "DROP INDEX $name IF EXISTS"

    drop_index_if_exists(driver, "my-index")

    driver.execute_query.assert_called_once_with(
        drop_query,
        {"name": "my-index"},
        database_=None,
    )


def test_create_fulltext_index_happy_path(driver: MagicMock) -> None:
    label = "node-label"
    text_node_properties = ["property-1", "property-2"]
    create_query = (
        "CREATE FULLTEXT INDEX $name IF NOT EXISTS "
        f"FOR (n:`{label}`) ON EACH "
        f"[{', '.join(['n.`' + property + '`' for property in text_node_properties])}]"
    )

    create_fulltext_index(driver, "my-index", label, text_node_properties)

    driver.execute_query.assert_called_once_with(
        create_query,
        {"name": "my-index"},
        database_=None,
    )


def test_create_fulltext_index_fail_if_exists(driver: MagicMock) -> None:
    label = "node-label"
    text_node_properties = ["property-1", "property-2"]
    create_query = (
        "CREATE FULLTEXT INDEX $name  "
        f"FOR (n:`{label}`) ON EACH "
        f"[{', '.join(['n.`' + property + '`' for property in text_node_properties])}]"
    )

    create_fulltext_index(
        driver, "my-index", label, text_node_properties, fail_if_exists=True
    )

    driver.execute_query.assert_called_once_with(
        create_query,
        {"name": "my-index"},
        database_=None,
    )


def test_create_fulltext_index_raises_error_with_neo4j_client_error(
    driver: MagicMock,
) -> None:
    label = "node-label"
    text_node_properties = ["property-1", "property-2"]
    driver.execute_query.side_effect = neo4j.exceptions.ClientError

    with pytest.raises(Neo4jIndexError):
        create_fulltext_index(driver, "my-index", label, text_node_properties)


def test_create_fulltext_index_empty_node_properties(driver: MagicMock) -> None:
    label = "node-label"
    node_properties: list[str] = []

    with pytest.raises(Neo4jIndexError) as excinfo:
        create_fulltext_index(driver, "my-index", label, node_properties)

    assert "Error for inputs to create_fulltext_index" in str(excinfo)


def test_create_fulltext_index_ensure_escaping(driver: MagicMock) -> None:
    label = "node-label"
    text_node_properties = ["property-1", "property-2"]
    create_query = (
        "CREATE FULLTEXT INDEX $name IF NOT EXISTS "
        f"FOR (n:`{label}`) ON EACH "
        f"[{', '.join(['n.`' + property + '`' for property in text_node_properties])}]"
    )

    create_fulltext_index(driver, "my-complicated-`-index", label, text_node_properties)

    driver.execute_query.assert_called_once_with(
        create_query,
        {"name": "my-complicated-`-index"},
        database_=None,
    )


def test_upsert_vector_happy_path(driver: MagicMock) -> None:
    id = 1
    embedding_property = "embedding"
    vector = [1.0, 2.0, 3.0]

    upsert_vector(driver, id, embedding_property, vector)

    upsert_query = (
        "MATCH (n) "
        "WHERE elementId(n) = $node_element_id "
        "WITH n "
        "CALL db.create.setNodeVectorProperty(n, $embedding_property, $vector) "
        "RETURN n"
    )

    driver.execute_query.assert_called_once_with(
        upsert_query,
        {
            "node_element_id": id,
            "embedding_property": embedding_property,
            "vector": vector,
        },
        database_=None,
    )


def test_upsert_vector_on_relationship_happy_path(driver: MagicMock) -> None:
    id = 1
    embedding_property = "embedding"
    vector = [1.0, 2.0, 3.0]

    upsert_vector_on_relationship(driver, id, embedding_property, vector)

    upsert_query = (
        "MATCH ()-[r]->() "
        "WHERE elementId(r) = $rel_element_id "
        "WITH r "
        "CALL db.create.setRelationshipVectorProperty(r, $embedding_property, $vector) "
        "RETURN r"
    )

    driver.execute_query.assert_called_once_with(
        upsert_query,
        {
            "rel_element_id": id,
            "embedding_property": embedding_property,
            "vector": vector,
        },
        database_=None,
    )


def test_upsert_vector_on_relationship_raises_neo4j_insertion_error(
    driver: MagicMock,
) -> None:
    id = 1
    embedding_property = "embedding"
    vector = [1.0, 2.0, 3.0]
    driver.execute_query.side_effect = neo4j.exceptions.ClientError

    with pytest.raises(Neo4jInsertionError) as excinfo:
        upsert_vector_on_relationship(driver, id, embedding_property, vector)
    assert "Upserting vector to Neo4j failed" in str(excinfo)


def test_upsert_vector_raises_neo4j_insertion_error(
    driver: MagicMock,
) -> None:
    id = 1
    embedding_property = "embedding"
    vector = [1.0, 2.0, 3.0]
    driver.execute_query.side_effect = neo4j.exceptions.ClientError

    with pytest.raises(Neo4jInsertionError) as excinfo:
        upsert_vector(driver, id, embedding_property, vector)

    assert "Upserting vector to Neo4j failed" in str(excinfo)


def test_upsert_vectors_wrong_entity_type(driver: MagicMock) -> None:
    with pytest.raises(ValueError) as exc_info:
        upsert_vectors(
            driver=driver,
            ids=["1"],
            embedding_property="embedding",
            embeddings=[[1.0, 2.0, 3.0]],
            neo4j_database="neo4j",
            entity_type="WRONG_ENTITY_TYPE",  # type: ignore[arg-type]
        )
    assert str(exc_info.value) == "entity_type must be either 'NODE' or 'RELATIONSHIP'"


def test_upsert_vectors_mismatched_ids_and_embeddings(driver: MagicMock) -> None:
    with pytest.raises(ValueError) as exc_info:
        upsert_vectors(
            driver=driver,
            ids=["1"],
            embedding_property="embedding",
            embeddings=[[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]],
            neo4j_database="neo4j",
        )
    assert str(exc_info.value) == "ids and embeddings must be the same length"


def test_upsert_vectors_inconsistent_embedding_sizes(driver: MagicMock) -> None:
    with pytest.raises(ValueError) as exc_info:
        upsert_vectors(
            driver=driver,
            ids=["1", "2"],
            embedding_property="embedding",
            embeddings=[[1.0, 2.0, 3.0], [4.0, 5.0]],
            neo4j_database="neo4j",
        )
    assert str(exc_info.value) == "All embeddings must be of the same size"



"""This example demonstrate how to embed a text into a vector
using MistralAI models and API.
"""

from neo4j_graphrag.embeddings import MistralAIEmbeddings

# set api key here on in the MISTRAL_API_KEY env var
api_key = None

embeder = MistralAIEmbeddings(model="mistral-embed", api_key=api_key)
res = embeder.embed_query("my question")
print(res[:10])



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import enum
import json
import logging
from pathlib import Path
from typing import Any, Optional, Union

import fsspec
import yaml
from fsspec.implementations.local import LocalFileSystem

logger = logging.getLogger(__name__)


class FileFormat(enum.Enum):
    JSON = "json"
    YAML = "yaml"

    @classmethod
    def json_valid_extension(cls) -> list[str]:
        return [".json"]

    @classmethod
    def yaml_valid_extension(cls) -> list[str]:
        return [".yaml", ".yml"]


class FileHandler:
    """Utility class to read/write JSON or YAML files.

    File format is guessed from the extension. Supported extensions are
    (lower or upper case):

    - .json
    - .yaml, .yml

    Example:

    .. code-block:: python

        from neo4j_graphrag.utils.file_handler import FileHandler
        handler = FileHandler()
        handler.read("my_file.json")

    If reading a file with a different extension but still in JSON or YAML format,
    it is possible to call directly the `read_json` or `read_yaml` methods:

    .. code-block:: python

        handler.read_yaml("my_file.txt")

    """

    def __init__(self, fs: Optional[fsspec.AbstractFileSystem] = None) -> None:
        self.fs = fs or LocalFileSystem()

    def _guess_file_format(self, path: Path) -> Optional[FileFormat]:
        # Note: .suffix returns an empty string if Path has no extension
        extension = path.suffix.lower()
        if extension in FileFormat.json_valid_extension():
            return FileFormat.JSON
        if extension in FileFormat.yaml_valid_extension():
            return FileFormat.YAML
        return None

    def _check_file_exists(self, path: Path) -> Path:
        if not path.exists():
            raise FileNotFoundError(f"File not found: {path}")
        return path

    def _read_json(self, path: Path) -> Any:
        """Reads a JSON file. If file does not exist, raises FileNotFoundError.

        Args:
            path (Path): The path of the JSON file.

        Raises:
             FileNotFoundError: If file does not exist.

        Returns:
            The parsed content of the JSON file.
        """
        logger.debug(f"FILE_HANDLER: read from json {path}")
        with self.fs.open(str(path), "r") as f:
            try:
                return json.load(f)
            except json.JSONDecodeError as e:
                raise ValueError("Invalid JSON file") from e

    def _read_yaml(self, path: Path) -> Any:
        """Reads a YAML file. If file does not exist, raises FileNotFoundError.

        Args:
            path (Path): The path of the YAML file.

        Raises:
             FileNotFoundError: If file does not exist.

        Returns:
            The parsed content of the YAML file.
        """
        logger.debug(f"FILE_HANDLER: read from yaml {path}")
        with self.fs.open(str(path), "r") as f:
            try:
                return yaml.safe_load(f)
            except yaml.YAMLError as e:
                raise ValueError("Invalid YAML file") from e

    def read(
        self, file_path: Union[Path, str], format: Optional[FileFormat] = None
    ) -> Any:
        """Try to infer file type from its extension and returns its
        parsed content.

        Args:
            file_path (Union[str, Path]): The path of the JSON file.
            format (Optional[FileFormat]): The file format to infer the file type from.
                If not set, the format is inferred from the extension.

        Raises:
            FileNotFoundError: If file does not exist.
            ValueError: If file extension is invalid or the file can not be parsed
                (e.g. invalid JSON or YAML)

        Returns: the parsed content of the file.
        """
        path = Path(file_path)
        path = self._check_file_exists(path)
        if not format:
            format = self._guess_file_format(path)
        if format == FileFormat.JSON:
            return self._read_json(path)
        if format == FileFormat.YAML:
            return self._read_yaml(path)
        raise ValueError(f"Unsupported file format: {format}")

    def _check_file_can_be_written(self, path: Path, overwrite: bool = False) -> None:
        """Check whether the file can be written to path with the following conditions:

        - If overwrite is set to True, file can always be written, any existing file will be overwritten.
        - If overwrite is set to False, and file already exists, file will not be overwritten.

        Args:
            path (Path): The path of the target file.
            overwrite (bool): If set to True, existing file will be overwritten. Default to False.

        Raises:
            ValueError: If file can not be written according to the above rules.
        """
        if overwrite:
            # we can overwrite, so no matter if file already exists or not
            return
        try:
            self._check_file_exists(path)
            # did not raise, meaning the file exists
            raise ValueError("File already exists. Use overwrite=True to overwrite.")
        except FileNotFoundError:
            # file not found all godo
            pass

    def _write_json(
        self,
        data: Any,
        file_path: Path,
        **extra_kwargs: Any,
    ) -> None:
        """Writes data to a JSON file

        Args:
            data (Any): The data to write.
            file_path (Path): The path of the JSON file.
            extra_kwargs (Any): Additional arguments passed to json.dump (e.g.: indent...). Note: a default indent=4 is applied.

        Raises:
            ValueError: If file can not be written according to the above rules.
        """
        fp = str(file_path)
        kwargs: dict[str, Any] = {
            "indent": 2,
        }
        kwargs.update(extra_kwargs)
        with self.fs.open(fp, "w") as f:
            json.dump(data, f, **kwargs)

    def _write_yaml(
        self,
        data: Any,
        file_path: Path,
        **extra_kwargs: Any,
    ) -> None:
        """Writes data to a YAML file

        Args:
            data (Any): The data to write.
            file_path (Path): The path of the YAML file.
            extra_kwargs (Any): Additional arguments passed to yaml.safe_dump. Note that we apply the following defaults:
                - "default_flow_style": False
                - "sort_keys": True
        """
        fp = str(file_path)
        kwargs: dict[str, Any] = {
            "default_flow_style": False,
            "sort_keys": True,
        }
        kwargs.update(extra_kwargs)
        with self.fs.open(fp, "w") as f:
            yaml.safe_dump(data, f, **kwargs)

    def write(
        self,
        data: Any,
        file_path: Union[Path, str],
        overwrite: bool = False,
        format: Optional[FileFormat] = None,
        **extra_kwargs: Any,
    ) -> None:
        """Guess file type and write it."""
        path = Path(file_path)
        self._check_file_can_be_written(path, overwrite)
        if not format:
            format = self._guess_file_format(path)
        if format == FileFormat.JSON:
            return self._write_json(data, path, **extra_kwargs)
        if format == FileFormat.YAML:
            return self._write_yaml(data, path, **extra_kwargs)
        raise ValueError(f"Unsupported file format: {format}")
